#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os
import sys
import time
import types
import itertools

import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt

from PIL import Image
from pandas import DataFrame, read_table
from pandas import __version__ as pandas_version
from distutils.version import LooseVersion

from scipy.io import whosmat, loadmat
from scipy.signal import convolve


# Constants to facilitate specifying groups of Dependent Variables
DV_FIXATED = ['fixated', 'count']
DV_FIXDUR = ['first', 'single', 'gaze', 'tofirst', 'total']
DV_ALL = ['fixated', 'count', 'fixid', 'first', 'single', 'gaze', 'tofirst', 'total', 'passes', 'refix']


class ImageSet(object):
    """ Set of images of equal size for masking and Feature creation. 

    Attributes:
        info (DataFrame): table of image metadata (filenames, size, type...)
        imageids (list): All unique image IDs in the set
        label (str): optional label to distinguish between ImageSets
        mat_var (str): name of MATLAB variable name if imported from .mat
        mat_contents (list): list of all variable names in .mat if applicable
        normalize (boolean): True if image data was normalized to 0..1 range
        preload (boolean): True if images were preloaded into memory
        size (tuple): image dimensions, specified as (width, height)
    """

    def __init__(self, images, mat_var=None, size=None, imageids=None, 
                 sep='\t', label=None, normalize=None, norm_range=None, preload=False):
        """ Create a new ImageSet and add specified images 

        Args:
            images: one of the following:
                - path to a single image or .mat file
                - path to a folder containing image or .mat files
                - list of image or .mat file names
                - simple text file, one filename per line
                - text / CSV file containing columns 'filename' and 'imageid'
            mat_var (str): variable to use if _images_ is a MATLAB file. If not specified,
                gridfix attempts to load the first variable in alphabetical order.
            size (tuple): image dimensions, specified as (width, height) in pixels
            imageids (list): string ID labels, one for each image. If not specified,
                file names without extension or a numerical label 0..n will be used
            sep (str): if _images_ is a text file, use this separator
            label (string): optional descriptive label
            normalize (boolean): normalize image color / luminance values to 0...1
                (defaults to True for images, False for MATLAB data)
            norm_range (tuple): normalization range. Defaults to (0, 255) for image
                files and per-image (min, max) for data read from MATLAB files
            preload (boolean): if True, load all images at creation time
                (faster, but uses a lot of memory)
        """
        self.imageids = []
        self._images = {}

        # DF to hold image metadata
        df_col = ['imageid', 'filename', 'width', 'height', 'channels', 'type', 'mat_var']
        self.info = DataFrame(columns=df_col)

        self.size = None
        self.label = label
        self.normalize = normalize
        self.norm_range = norm_range
        self.preload = preload
        self._last_image = None
        self._last_imageid = None

        self.mat_var = None
        self.mat_contents = None

        if size is not None:
            self.size = tuple(size)

        self._add_images(images, mat_var, imageids, sep)


    def __repr__(self):
        """ Short string representation for printing """
        # Image size
        if self.size is None:
            size = 'undefined'
        else:
            size = str(self.size)

        # Number of images
        s = ''
        if len(self.imageids) != 1:
            s = 's'

        desc = ''
        if self.label is not None:
            desc = ' "{:s}"'.format(str(self.label))

        mat = ''
        if self.mat_var is not None:
            mat = ', mat_var={:s}'.format(self.mat_var)
        norm = ''
        if self.normalize:
            norm = ', normalized'
        r = '<gridfix.ImageSet{:s}, {:d} image{:s}, size={:s}{:s}{:s}>'
        return r.format(desc, len(self.imageids), s, size, mat, norm)


    def __len__(self):
        """ Overload len(ImageSet) to report number of images. """
        return len(self.imageids)


    def __iter__(self):
        """ Overload iteration to step through the ndarray representations of images. """
        return iter([np.array(self._images[i]) for i in self._images.keys()])


    def __getitem__(self, imageid):
        """ Allow to retrieve image by bracket indexing """
        return self.image(imageid)


    def _add_images(self, images, mat_var=None, imageids=None, sep='\t'):
        """ Add one or more image(s) to set. 

        Args:
            images: one of the following: 
                - path to a single image file
                - path to a folder containing image files
                - list of image file names
                - simple text file, one image filename per line
                - text / CSV file containing columns 'filename' and 'imageid'
            mat_var (str): variable to use if _images_ is a MATLAB file
            imageids (list): string ID labels, one for each image. If not specified,
                file names without extension or a numerical label 0..n will be used
            sep (str): if _images_ is a text file, use this separator
        """
        filelist = []
        if imageids is None:
            imageids = []

        img_root = os.getcwd()

        # Build file list
        if type(images) == list:
            filelist = images
        
        elif type(images) == str:

            if os.path.isdir(images):
                # Directory
                filelist = [os.path.join(images, b) for b in sorted(os.listdir(images))]

            elif os.path.isfile(images):
                # Single file - check if it is a text list of images!
                (ifname, ifext) = os.path.splitext(images)
                
                if ifext.lower() in ['.txt', '.tsv', '.csv', '.dat']:
                    # assume image list
                    try:
                        imgfiles = read_table(images, header=None, index_col=False, sep=sep)

                        if imgfiles.shape[1] == 1:
                            # Only one column: assume no headers and load as list of filenames
                            filelist = list(imgfiles.iloc[:, 0])

                        elif imgfiles.shape[1] > 1:
                            # More than one column: look for 'imageid' and 'filename' columns
                            if 'imageid' in list(imgfiles.iloc[0, :]) and 'filename' in list(imgfiles.iloc[0, :]):
                                imgfiles.columns = imgfiles.iloc[0, :]
                                imgfiles = imgfiles.iloc[1:, :]
                                filelist = list(imgfiles['filename'])
                                imageids = list(imgfiles['imageid'])

                        lfolder, lname = os.path.split(images)
                        if len(lfolder) > 0:
                            img_root = os.path.abspath(lfolder)

                    except: 
                        raise ValueError('could not read image list file, check format!')
                else:
                    # assume single image file
                    filelist = [images]
        else:
            raise ValueError('first argument must be list or a string containing a file or directory!')

        # Verify image files
        filetable = []
        for (idx, ifile) in enumerate(filelist):
            try:
                imeta = self._verify_image(ifile, mat_var, img_root)

                # assign imageid
                (ifdir, iffile) = os.path.split(ifile)
                (ifbase, ifext) = os.path.splitext(iffile)
                if imageids is not None and len(imageids) > 0:
                    imageid = imageids.pop(0)
                else:
                    imageid = ifbase
                imeta['imageid'] = imageid

                if imageid in self.imageids:
                    print('Warning: replacing existing image with ID <{:s}>!'.format(imageid))
                else:
                    self.imageids.append(imageid)

                filetable.append(imeta)

            except ValueError as err:
                print('Warning: file {:s} could not be added, error: {:s}!'.format(ifile, str(err)))

        df_col = ['imageid', 'filename', 'width', 'height', 'channels', 'type', 'mat_var']
        self.info = DataFrame(filetable, columns=df_col)
        self.imageids = list(self.info.imageid)

        # Preload files if requested
        for imid in self.imageids:
            if self.preload:
                self._images[imid] = self._load_image(imid)
            else:
                self._images[imid] = None


    def image(self, imageid):
        """ Get image by imageid, loading it if not preloaded

        Args:
            imageid (str): valid imageid from set

        Returns:
            ndarray of raw image data
        """
        if imageid not in self.imageids:
            raise ValueError('Specified _imageid_ does not exist!')

        if self._images[imageid] is not None:
            return self._images[imageid]
        else:
            if self._last_imageid == imageid:
                return self._last_image
            else:
                return self._load_image(imageid)


    def _load_image(self, imageid):
        """ Load and return image data for imageid. 

        Args:
            imageid (str): valid imageid from set

        Returns:
            ndarray of raw image data
        """
        if imageid not in self.imageids:
            raise ValueError('Specified _imageid_ does not exist!')

        if self._last_imageid is not None and self._last_imageid == imageid:
            return self._last_image

        imdata = None
        imeta = self.info[self.info.imageid == imageid]

        if len(imeta) > 0:

            if imeta.type.iloc[0] == 'MAT':
                mat = loadmat(imeta.filename.iloc[0], variable_names=imeta.mat_var.iloc[0])
                imdata = np.asarray(mat[imeta.mat_var.iloc[0]], dtype=float)

            else:
                i = Image.open(imeta.filename.iloc[0])
                if i.mode == 'RGBA':
                    i = i.convert('RGB')    # Drop alpha channel
                imdata = np.asarray(i, dtype=float)

            if self.normalize is None and imeta.type.iloc[0] != 'MAT':
                if self.norm_range is None:
                    imdata = imdata / 255.0
                else:
                    imdata = (imdata - self.norm_range[0]) / (self.norm_range[1] - self.norm_range[0])
                self.normalize = True

            elif self.normalize:
                if self.norm_range is None:
                    imdata = (imdata - imdata.min()) / (imdata.max() - imdata.min())
                else:
                    imdata = (imdata - self.norm_range[0]) / (self.norm_range[1] - self.norm_range[0])

            self._last_imageid = imageid
            self._last_image = imdata

        return imdata


    def _verify_image(self, image_file, mat_var=None, img_root=None):
        """ Verify type and size of image without actually loading data

        Args: 
            image_file (str): path to image file to verify
            mat_var (str): optional variable name for MATLAB files
            img_root (str): folder containing image list in case file paths are relative

        Returns:
            dict of metadata, column names as in imageset.info DataFrame
        """
        if not os.path.isfile(image_file):
            if img_root is not None and os.path.isfile(os.path.join(img_root, image_file)):
                image_file = os.path.join(os.path.join(img_root, image_file))
            else:
                raise ValueError('file not found')

        (ifbase, ifext) = os.path.splitext(image_file)

        imeta = {'imageid': None, 
                 'filename': '',
                 'width': -1,
                 'height': -1, 
                 'channels': -1, 
                 'type': None,
                 'mat_var': None}
        
        # Matlab files
        if ifext.lower() == '.mat':
            try:
                # Load .mat header and identify variables
                tmp = whosmat(image_file)
                tmpvars = [m[0] for m in tmp]
                
                if self.mat_contents is None:
                    self.mat_contents = tmpvars
                
                if mat_var is None: 
                    mat_var = tmpvars[0]

                if mat_var in tmpvars:
                    
                    if self.mat_var is None:
                        self.mat_var = mat_var

                    # check image size
                    imshape = [m[1] for m in tmp if m[0] == mat_var][0]
                    if self.size is None:
                        self.size = (imshape[1], imshape[0])

                    if (imshape[1], imshape[0]) == self.size:
                        imeta['filename'] = image_file
                        imeta['width'] = imshape[1]
                        imeta['height'] = imshape[0]
                        if len(imshape) > 2:
                            imeta['channels'] = imshape[2]
                        else:
                            imeta['channels'] = 1
                        imeta['type'] = 'MAT'
                        imeta['mat_var'] = mat_var
                    
                    else:
                        w = 'Warning: skipping {:s} due to image size ({:d}x{:d} instead of {:d}x{:d}).'
                        print(w.format(iffile, imshape[1], imshape[0], self.size[0], self.size[1]))

                else:
                    raise ValueError('specified MATLAB variable not in file')

            except: 
                raise ValueError('error loading MATLAB data')

        # Image files
        else:
            try:
                i = Image.open(image_file) 
                imsize = i.size

                if self.size is None:
                    self.size = imsize

                # check image size
                if imsize == self.size:
                    imeta['filename'] = image_file
                    imeta['width'] = imsize[0]
                    imeta['height'] = imsize[1]
                    if i.mode in ['RGB', 'RGBA']:
                        imeta['channels'] = 3
                    else:
                        imeta['channels'] = 1
                    imeta['type'] = i.format
                    imeta['mat_var'] = ''
                    
                else:
                    w = 'Warning: skipping {:s} due to image size ({:d}x{:d} instead of {:d}x{:d}).'
                    print(w.format(image_file, imsize[0], imsize[1], self.size[0], self.size[1]))

            except OSError:
                raise ValueError('not an image or file could not be opened.')

        return imeta


    def plot(self, imageid, cmap='gray', image_only=False, ax=None):
        """ Display one of the contained images by imageid using matplotlib

        Args:
            imageid (str): valid imageid of image to show
            cmap (str): name of a matplotlib colormap to use
            image_only (boolean): if True, return only image content without labels
            ax (Axes): axes object to draw on, to include result in other figure

        Returns:
            matplotlib figure object, or None if passed an axis to draw on
        """
        try:
            if ax is not None:
                ax1 = ax
            else:
                fig = plt.figure()
                ax1 = fig.add_subplot(1,1,1)

            ax1.imshow(self.image(imageid), cmap=plt.get_cmap(cmap))
            if image_only:
                ax1.axis('off')
            else:
                ax1.set_title(imageid)

            if ax is None and not plt.isinteractive():
                # Only return figure object in non-interactive mode, otherwise
                # IPython/Jupyter will display the figure twice (once while plotting
                # and once as the cell result)!
                return fig

        except KeyError:
            raise ValueError('No image with ID "{:s}" in set'.format(imageid))


class Fixations(object):
    """ Dataset of fixation locations.

    Fixation locations are assumed to be one-indexed in input, e.g. 1..800 if 
    the image is 800 pixels wide, and converted internally to Pythons zero-indexed
    array convention.

    Attributes:
        data (DataFrame): DataFrame of raw fixation data
        has_times (boolean): True if fixation times have been loaded
        imageids (list): all unique image IDs represented in the dataset
        imageset (ImageSet): if present, the associated ImageSet
        input_file (str): file name of fixation data file
        num_samples (int): number of fixation samples
        num_vars (int): number of columns / variables in dataset 
        offset (tuple): optional offset from raw (x,y) positions
        shape (tuple): dimensions of .data, i.e. same as Fixations.data.shape
        variables (list): list of all variables loaded from input file
    """

    def __init__(self, data, imageset=None, offset=(0, 0), imageid='imageid', 
                 fixid='fixid', x='x', y='y', fixstart=None, fixend=None,
                 fixdur=None, numericid=False, round_coords=True):
        """ Create new Fixations dataset and calculate defaults.

        Args:
            data: a predefined DataFrame or name of a file containing fixation report
            imageset (ImageSet): if present, verify imageids against this ImageSet
            offset (tuple): an (x, y) tuple of pixel values to offset fixations. E.g.,
                if fixations have their origin at image center, use (-width/2, -height/2)
            imageid (str): name of data file column containing imageids
            fixid (str): name of data file column with unique fixation ID / index
            x (str): name of data file column for horizontal fixation locations
            y (str): name of data file column for vertical fixation locations
            fixstart (str): name of data file column containing fixation start time
            fixend (str): name of data file column containing fixation end time
            fixdur (start): name of column containing fixation durations. If start/end
                are given but this is None, durations will be calculated by gridfix
            numericid (boolean): if True, try to force parsing imageid as numeric
            round_coords (boolean): How to treat float values in fixation coordinates:
                if True, round values to closest pixel (e.g., x=4.4 -> pixel 4 / index 3)
                if False, round values up to next integer (x=4.4 -> pixel 5 / index 4)
        """
        self.data = DataFrame()

        if isinstance(data, DataFrame):
            # data is already a DataFrame
            self.data = data
            self.input_file = None

        else:
            try:
                self.data = read_table(data, index_col=False)
                self.input_file = data

            except IOError:
                raise IOError('Could not load fixation data, check file name and type!')

        # Internal column names
        self._imageid = imageid
        self._fixid = fixid
        self._x = x
        self._y = y
        self._xpx = self._x + '_PX'
        self._ypx = self._y + '_PX'
        self._fixstart = fixstart
        self._fixend = fixend
        self._fixdur = fixdur

        # Verify that all required columns are present
        cols = [imageid, fixid, x, y]
        miss_cols = []
        for c in cols:
            if c not in self.data.columns.values:
                miss_cols.append(c)

        if len(miss_cols) > 0:
            raise ValueError('Missing columns ({:s}), please specify column names!'.format(str(miss_cols)))
            
        # Image ID column should always be converted to strings
        if numericid:
            self.data[self._imageid] = self.data[self._imageid].astype(int).astype(str)
        else:
            self.data[self._imageid] = self.data[self._imageid].astype(str)

        self.imageids = list(self.data[self._imageid].unique())
        self.shape = self.data.shape
        self.num_samples = self.shape[0]
        self.num_vars = self.shape[1]
        self.variables = list(self.data.columns.values)

        # Fixation timing columns (optional, default: not specified)
        self.has_times = False
        if fixstart is not None and fixend is not None:
            if fixstart not in self.data.columns.values:
                raise ValueError('Unknown column specified for fixation start time: "{:s}"'.format(fixstart))
            if fixend not in self.data.columns.values:
                raise ValueError('Unknown column specified for fixation end time: "{:s}"'.format(fixend))
            self.has_times = True

            # Calculate fixation durations if no pre-computed column was specified
            if self._fixdur is None:
                self._fixdur = '__FIXDUR'
                self.data[self._fixdur] = self.data[self._fixend] - self.data[self._fixstart]
        else:
            if (fixstart is None) - (fixend is None) != 0:
                raise ValueError('Optional timing columns (fixstart, fixend) must be specified together!')

        # If ImageSet provided, check if all images are present
        self.imageset = None
        if imageset is not None:
            self.imageset = imageset
            missing_imgs = []
            for im in self.imageids:
                if im not in imageset.imageids:
                    missing_imgs.append(im)
            if len(missing_imgs) > 0:
                print('Warning: the following images appear in the fixation data but not the speficied ImageSet: {:s}'.format(', '.join(missing_imgs)))

        # Set offset and calculate pixel indices (_xpx, _ypx)
        self.offset = (0, 0)
        self.round_coords = round_coords
        self.set_offset(offset)
        

    def __repr__(self):
        """ String representation """
        r = '<gridfix.Fixations data set, {:d} samples, {:d} images>'.format(self.num_samples, len(self.imageids))
        if self.imageset is not None:
            r += '\nImages:\n\t{:s}'.format(str(self.imageset))
        return r


    def __len__(self):
        """ Overload len() to report the number of samples """
        return self.data.shape[0]


    def __getitem__(self, imageid):
        """ Bracket indexing returns all fixations for a specified image """
        return self.select_fix(select={self._imageid: imageid})


    def set_offset(self, offset):
        """ Set a constant offset for eye x/y coordinates.

        If image coordinates are relative to image center, use (-width/2, -height/2)
        (GridFix uses a coordinate origin at the top left).

        Args:
            offset (tuple): 2-tuple of (hor, ver) offset values in pixels
        """
        # Reset previous offset
        prevoffset = self.offset
        if prevoffset[0] != 0.0 or prevoffset[1] != 0.0:
            self.data[self._x] = self.data[self._x] - prevoffset[0]
            self.data[self._y] = self.data[self._y] - prevoffset[1]

        self.data[self._x] = self.data[self._x] + offset[0]
        self.data[self._y] = self.data[self._y] + offset[1]
        self.offset = (offset[0], offset[1])

        # Convert x/y fixation positions to integers (pixels) while keeping original data
        # Note: we're going to use these as indices and Python is 0-indexed, so subtract 1!
        if self.round_coords:
            self.data[self._xpx] = np.asarray(np.round(self.data[self._x]), dtype=int) - 1
            self.data[self._ypx] = np.asarray(np.round(self.data[self._y]), dtype=int) - 1
        else:
            self.data[self._xpx] = np.asarray(np.ceil(self.data[self._x]), dtype=int) - 1
            self.data[self._ypx] = np.asarray(np.ceil(self.data[self._y]), dtype=int) - 1


    def select_fix(self, select={}):
        """ Return a subset of fixation data for specified imageid.

        Args:.
            select (dict): dict of filter variables, as {column: value}

        Returns:
            New Fixations object containing selected fixations only
        """
        if self._imageid not in select.keys():
            print('Warning: no image ID in filter variables, selection will yield fixations from multiple images! Proceed at own risk.')

        if select[self._imageid] not in self.data[self._imageid].values:
            print('Warning: zero fixations selected for specified imageid ({:s})'.format(select[self._imageid]))

        selection = self.data
        if len(select) > 0:
            for col, target in select.items():
                if col not in selection.columns.values:
                    print('Warning: filter variable {:s} not found in Fixations dataset!'.format(col))
                else:
                    # Make sure dict value is list-like
                    if type(target) not in (tuple, list):
                        target = [target]
                    selection = selection[selection[col].isin(target)]

        result = Fixations(selection.copy(), imageid=self._imageid, fixid=self._fixid,
                           x=self._x, y=self._y, imageset=self.imageset,
                           fixstart=self._fixstart, fixend=self._fixend, fixdur=self._fixdur,
                           offset=self.offset, round_coords=self.round_coords)
        return result


    def plot(self, imageid=None, select={}, on_image=True, oob=False,
             plotformat='wo', plotsize=5.0, plotcolor=[1, 1, 1],
             durations=False, image_only=False, ax=None):
        """ Plot fixations for selected imageid, either alone or on image

        Args:
            imageid (str): optional image ID to plot fixations for
            select (dict): dict of additional filter variables (see select_fix())
            image (bool): if True, superimpose fixations onto image (if ImageSet present)
            oob (bool): if True, include out-of-bounds fixations when plotting
            plotformat (str): format string for plt.pyplot.plot(), default: white circles
            plotsize (float): fixation marker size
            plotcolor (color): matplotlib color spec for markers
            durations (bool): if True, plot duration of each fixation next to marker
            image_only (boolean): if True, return only image content without labels
            ax (Axes): axes object to draw to, to include result in other figure

        Returns:
            matplotlib figure object, or None if passed an axis to draw on
        """
        if imageid is not None:
            if imageid not in select.keys():
                select[self._imageid] = imageid
            plotfix = self.select_fix(select)
        else:
            plotfix = self

        if ax is not None:
            ax1 = ax 
        else:
            fig = plt.figure()
            ax1 = fig.add_subplot(1,1,1)

        try:
            if on_image:
                ax1.imshow(self.imageset.image(imageid), origin='upper')

        except AttributeError:
            print('Warning: cannot view fixations on image due to missing ImageSet!')

        if oob:
            ax1.plot(plotfix.data[self._xpx], plotfix.data[self._ypx], plotformat,
                     markersize=plotsize, color=plotcolor)
            if durations and self._fixdur in plotfix.data.columns.to_list():
                for r in plotfix.data.iterrows():
                    x = r[1][self._xpx]
                    if r[1][self._ypx] > 15:
                        y = r[1][self._ypx] - 15
                    else:
                        y = r[1][self._ypx] + 5
                    d = r[1][self._fixdur]
                    ax1.text(x, y, str(d), horizontalalignment='center')

        else:
            try:
                if self.imageset is not None:
                    size = self.imageset.size
                else:
                    size = (max(plotfix.data[self._xpx]), max(plotfix.data[self._ypx]))
                fix = plotfix.data[(plotfix.data[self._xpx] >= 0) &
                                   (plotfix.data[self._xpx] < size[0]) &
                                   (plotfix.data[self._ypx] >= 0) &
                                   (plotfix.data[self._ypx] < size[1])]
                ax1.plot(fix[self._xpx], fix[self._ypx], plotformat,
                     markersize=plotsize, color=plotcolor)
                if durations and self._fixdur in plotfix.data.columns.to_list():
                    for r in fix.iterrows():
                        x = r[1][self._xpx]
                        if r[1][self._ypx] > 15:
                            y = r[1][self._ypx] - 15
                        else:
                            y = r[1][self._ypx] + 5
                        d = r[1][self._fixdur]
                        ax1.text(x, y, str(d), horizontalalignment='center')

                ax1.set_xlim((0, size[0]))
                ax1.set_ylim((0,size[1]))
                ax1.invert_yaxis()
                if image_only:
                    ax1.axis('off')
                else:
                    ax1.set_title(imageid)

            except AttributeError:
                print('Warning: cannot filter fixations for image boundaries due to missing ImageSet!')
                ax1.plot(plotfix[self._xpx], plotfix[self._ypx], plotformat,
                     markersize=plotsize, color=plotcolor)
                ax1.invert_yaxis()

        if ax is None and not plt.isinteractive():  # see ImageSet.plot()
            return fig


    def location_map(self, imageid=None, size=None):
        """ Binary ndarray of fixated and non-fixated pixels within image area

        Args:
            imageid (str): optional image ID to create map for one image only
            size (tuple): image dimensions, specified as (width, height).

        Returns:
            2d boolean ndarray, True where fixated, otherwise False
        """
        if imageid is not None:
            mapfix = self.select_fix({self._imageid: imageid})
        else:
            mapfix = self

        if size is None:
            if self.imageset is not None:
                size = self.imageset.size
            else:
                raise ValueError('Image size or attached ImageSet are necessary for location mapping!')

        fixloc = np.zeros((size[1], size[0]), dtype=bool)
        fix = mapfix.data[(self.data[self._xpx] >= 0) & (self.data[self._xpx] < size[0]) &
                          (self.data[self._ypx] >= 0) & (self.data[self._ypx] < size[1])]
        fixloc[fix[self._ypx], fix[self._xpx]] = True
        return fixloc


    def count_map(self, imageid=None, size=None):
        """ Map of fixation counts for each image pixel

        Args:
            imageid (str): optional image ID to create map for one image only
            size (tuple): image dimensions, specified as (width, height).

        Returns:
            2d ndarray of pixel fixation counts
        """
        if imageid is not None:
            mapfix = self.select_fix({self._imageid: imageid})
        else:
            mapfix = self

        if size is None:
            if self.imageset is not None:
                size = self.imageset.size
            else:
                raise ValueError('Image size or attached ImageSet are necessary for location mapping!')

        fixcount = np.zeros((size[1], size[0]), dtype=int)
        fix = mapfix.data[(self.data[self._xpx] >= 0) & (self.data[self._xpx] < size[0]) &
                          (self.data[self._ypx] >= 0) & (self.data[self._ypx] < size[1])]
        fixloc = fix[[self._ypx, self._xpx]].to_numpy()
        for pos in fixloc:
            fixcount[pos[0], pos[1]] += 1
        return fixcount



    def dur_map(self, imageid=None, size=None):
        """ Map of total fixation durations for each image pixel

        Args:
            imageid (str): optional image ID to create map for one image only
            size (tuple): image dimensions, specified as (width, height).

        Returns:
            2d ndarray of fixation durations at each pixel
        """

        if imageid is not None:
            mapfix = self.select_fix({self._imageid: imageid})
        else:
            mapfix = self

        if not self._fixdur in mapfix.data.columns.to_list():
            raise ValueError('Duration map is only available if dataset contains fixation durations!')


        if size is None:
            if self.imageset is not None:
                size = self.imageset.size
            else:
                raise ValueError('Image size or attached ImageSet are necessary for location mapping!')

        durmap = np.zeros((size[1], size[0]), dtype=float)
        fix = mapfix.data[(self.data[self._xpx] >= 0) & (self.data[self._xpx] < size[0]) &
                  (self.data[self._ypx] >= 0) & (self.data[self._ypx] < size[1])]
        for r in fix.iterrows():
            durmap[r[1][self._ypx], r[1][self._xpx]] += r[1][self._fixdur]
        return durmap



    def _gauss2d(self, size_x=100, sigma_x=10, size_y=None, sigma_y=None):
        """ Create 2D Gaussian kernel for heat map visualizations

        Args:
            size_x, size_y (int): horizontal/vertical kernel size, in pixels
            sigma_x, sigma_y (float): hor/vert kernel standard deviaton, in pixels

        Returns:
            2d ndarray containing Gaussian kernel
        """
        if size_y is None:
            size_y = size_x
        if sigma_y is None:
            sigma_y = sigma_x

        (xx, yy) = np.meshgrid(np.linspace(-size_x/2, size_x/2, size_x), np.linspace(-size_y/2, size_y/2, size_y))
        G = np.exp(-1.0*(((xx*xx) / (2*sigma_x*sigma_x)) + ((yy*yy) / (2*sigma_y*sigma_y))))
        return G



    def heat_map(self, imageid=None, size=None, dur=False, convolution=True,
                 size_x=100, sigma_x=10, size_y=None, sigma_y=None, normalize=False, threshold=None):
        """ 2D heat map using convolution with a Gaussian kernel.
            Best for fixations or samples in screen pixel coordinates.

        Args:
            imageid (str): optional image ID to create map for one image only
            size (tuple): image dimensions, specified as (width, height).
            convolution (bool): if False, use a sparse addition approch to calculate
                heat map. Works if scipy is unavailable, faster if very few fixations
            dur (bool): if True, weight heat map by fixation durations if available
            size_x, size_y (int): horizontal/vertical kernel size, in pixels
            sigma_x, sigma_y (float): hor/vert kernel standard deviaton, in pixels
            normalize (bool): if True, return values in range 0..1
            threshold (float): threshold map by masking output array at a given value
        """
        if size_y is None:
            size_y = size_x
        if sigma_y is None:
            sigma_y = sigma_x

        if dur and not self._fixdur in self.data.columns.to_list():
            print('Warning: no fixation durations in dataset, cannot apply duration weighting (dur=True)')
            dur = False

        kernel = self._gauss2d(size_x, sigma_x, size_y, sigma_y)

        if not convolution:
            # Use manual sparse addition of kernels
            if imageid is not None:
                mapfix = self.select_fix({self._imageid: imageid})
            else:
                mapfix = self

            if size is None:
                if self.imageset is not None:
                    size = self.imageset.size
                else:
                    raise ValueError('Image size or attached ImageSet are necessary for location mapping!')

            hmap = np.zeros((size[1]+2*size_x, size[0]+2*size_y))
            hwx = int(round(size_x/2))
            hwy = int(round(size_y/2))

            fix = mapfix.data[(self.data[self._xpx] >= 0) & (self.data[self._xpx] < size[0]) &
                              (self.data[self._ypx] >= 0) & (self.data[self._ypx] < size[1])]
            for r in fix.iterrows():
                pos = (r[1][self._ypx], r[1][self._xpx])
                xl = pos[1]+size_x-hwx
                xr = pos[1]+size_x+hwx
                yl = pos[0]+size_y-hwy
                yr = pos[0]+size_y+hwy
                if dur:
                    hmap[yl:yr, xl:xr] += kernel * r[1][self._fixdur]
                else:
                    hmap[yl:yr, xl:xr] += kernel
            heatmap = hmap[size_y:size[1]+size_y, size_x:size[0]+size_x]

        else:
            # Use scipy.signal.convolve (default)
            if dur:
                hmap = self.dur_map(imageid, size).astype(np.float64)
            else:
                hmap = self.count_map(imageid, size).astype(np.float64)
            heatmap = convolve(hmap, kernel, mode='same')

        if normalize:
            heatmap = (heatmap-heatmap.min()) / (heatmap.max()-heatmap.min())
        if threshold is not None:
            heatmap = np.ma.masked_less_equal(heatmap, threshold)

        return heatmap



    def assign_regions(self, regionset, regionno=True, col_prefix=None):
        """ Assign each fixation the corresponding region from a RegionSet. New columns
        are added directly to the Fixation object's underlying DataFrame.

        Note: in case of overlapping regions, fixatiońs will be labeled as the higher
        region number (likely the region which appeared later in the RegionSet input file)!

        Args:
            regionset: RegionSet object to match against (must contain same imageids)
            regionno (bool): if True, add column for region number in addition to regionid string
            col_prefix (str): name prefix for new column names
        """

        if not regionset.is_global and not all(iid in regionset.imageids for iid in self.imageids):
            raise ValueError('At least one imageid was not found in the specified RegionSet!')

        # Generate new unique column names using RegionSet label if available
        idcol = 'regionid'
        rcol = 'regionno'

        if col_prefix is not None:
            idcol = '{:s}_{:s}'.format(col_prefix, idcol)
            rcol = '{:s}_{:s}'.format(col_prefix, rcol)
        else:
            if regionset.label is not None:
                idcol = '{:s}_{:s}'.format(regionset.label, idcol)
                rcol = '{:s}_{:s}'.format(regionset.label, rcol)

        if idcol in self.data.columns:
            idcol_num = idcol
            num = 1
            while idcol_num in self.data.columns:
                idcol_num = '{:s}{:d}'.format(idcol, num)
                num += 1
            idcol = idcol_num

        if regionno and rcol in self.data.columns:
            rcol_num = rcol
            num = 1
            while rcol_num in self.data.columns:
                rcol_num = '{:s}{:d}'.format(rcol, num)
                num += 1
            rcol = rcol_num

        self.data.loc[:, idcol] = 'none'
        if regionno:
            self.data[rcol] = np.nan

        # Filter for out-of-bounds fixations
        valid_fix = ((self.data[self._xpx] >= 0) & (self.data[self._xpx] < regionset.size[0]) &
                    (self.data[self._ypx] >= 0) & (self.data[self._ypx] < regionset.size[1]))

        if regionset.is_global:
            # Apply global region map to all fixation coordinates
            rmap = regionset.region_map(imageid='*', ignore_background=False)

            rnos = np.ones(self.data.shape[0]) * np.nan
            rnos[valid_fix] = rmap[self.data.loc[valid_fix, self._ypx], self.data.loc[valid_fix,self._xpx]]
            if regionno:
                self.data.loc[:, rcol] = rnos

            # Resolve regionids from DataFrame (not guaranteed to be sorted by regionno)
            for rn in np.unique(rnos[~np.isnan(rnos)]):
                self.data.loc[rnos == rn, idcol] = regionset.info.regionid[regionset.info.regionno == rn].values

        else:
            # Resolve regions for each imageid
            for imid in self.imageids:
                img_fix = valid_fix & (self.data[self._imageid] == imid)
                rmap = regionset.region_map(imageid=imid, ignore_background=False)
                rnos = np.ones(self.data.shape[0]) * np.nan
                rnos[img_fix] = rmap[self.data.loc[img_fix, self._ypx], self.data.loc[img_fix,self._xpx]]

                region_info = regionset.info[regionset.info.imageid == imid]
                if regionno:
                    self.data.loc[img_fix, rcol] = rnos[img_fix]
                for rn in np.unique(rnos[~np.isnan(rnos)]):
                    self.data.loc[img_fix & (rnos == rn), idcol] = region_info.regionid[region_info.regionno == rn].values



class FixationModel(object):
    """ Combines Features and Fixations to create predictors and R source for GLMM

    Attributes:
        chunks (list): list of data columns that define chunks (e.g., subjects or sessions)
        comp_features (dict): dict of labelled feature comparisons. Predictors will be replicated
            for each feature in a comparison so that features can serve as fixed or random factor
        exclude_first_fix (boolean): if True, first fixation index was set NaN (e.g., fixation cross)
        exclude_last_fix (boolean): if True, last fixation index was set NaN (e.g., image offset)
        features (dict): dictionary of feature objects and feature groups
        predictors (DataFrame): model predictors for GLMM
        regionset (RegionSet): attached RegionSet
        runtime (float): time in seconds for most recent update of predictor matrix
        normalize_features (bool): if True, all feature values are normalized to 0..1 range
    """

    def __init__(self, fixations, regionset, dv_type='fixated', features=None, feature_labels=None,
                 chunks=[], progress=False, exclude_first_fix=False, exclude_last_fix=False,
                 normalize_features=None):
        """ Create a new FixationModel.

        Args:
            fixations (Fixations): fixation data to use as model DV (column 'fixation')
            regionset (RegionSet): a RegionSet object defining length of all features
            dv_type (str): type of DV to generate, or list of multiple options:
                'fixated': binary coding of fixated (1) and unfixated (0) regions
                'count': absolute fixation count for each region
                'fixid': fixation ID of the first valid fixation within each region (NaN if not fixated)
                'passes': total number of viewing passes on this object (NaN if not fixated)
                'refix': total number of refixations (viewing passes after first pass; NaN if not fixated)
                'total': total fixation time for each region (NaN if not fixated)
                'first': first fixation duration per region (NaN if not fixated)
                'gaze': first-pass gaze duration per region (all initial fixations without refixations)
                'single': fixation duration if region was fixated exactly once (NaN if not fixated)
                'tofirst': start time of the first fixation on each region (NaN if not fixated)
            features (list): list of Feature objects to add (use add_comparison for feature groups)
            feature_labels (list): string labels to apply to features defined using features= attribute
            chunks (list): list of fixation data columns that define chunks (e.g., subjects or sessions)
            progress (bool): print current image and group variables to indicate model build progress
            exclude_first_fix (bool): if True, set first fixated region per image to NaN for GLMM
            exclude_last_fix (str): controls how to deal with regions receiving the last image fixation:
                'never' or False: do not handle the last fixation specially
                'always' or True: drop the entire region if it received the last fixation at any time
                'pass': exclude viewing pass (one or multiple fixations) that received the last fixation
            normalize_features (bool): if True, normalize all feature values to 0..1 range
        """
        self.regionset = regionset

        self._fix = fixations
        self._pred = DataFrame()
        self._consistent = False    # Flags whether we need to rebuild predictor matrix

        self.features = {}
        self.comp_features = {}
        self.normalize_features = normalize_features
        self.exclude_first_fix = exclude_first_fix
        self.exclude_last_fix = exclude_last_fix

        # Supported DVs: labels and other model parameters
        self._dvs = {'fixated': {'f':'c', 'rvar': 'dvFix',    'fun': 'glmer', 'family': 'binomial'},
                     'count':   {'f':'c', 'rvar': 'dvCount',  'fun': 'glmer', 'family': 'poisson'},
                     'fixid':   {'f':'c', 'rvar': 'fixID',    'fun': None,    'family': None},
                     'passes':  {'f':'c', 'rvar': 'dvPasses', 'fun': 'glmer', 'family': 'poisson'},
                     'refix':   {'f':'c', 'rvar': 'dvRefix',  'fun': 'glmer', 'family': 'poisson'},
                     'first':   {'f':'t', 'rvar': 'dvFirst',  'fun': 'lmer',  'family': None},
                     'gaze':    {'f':'t', 'rvar': 'dvGaze',   'fun': 'lmer',  'family': None},
                     'tofirst': {'f':'t', 'rvar': 'dvToFirst','fun': 'lmer',  'family': None},
                     'total':   {'f':'t', 'rvar': 'dvTotal',  'fun': 'lmer',  'family': None},
                     'single':  {'f':'t', 'rvar': 'dvSingle', 'fun': 'lmer',  'family': None}
                    }

        if type(dv_type) != list:
            dv_type = [dv_type,]
        for dvt in dv_type:
            if dvt not in self._dvs.keys():
                raise ValueError('Error: unknown DV type specified: "{:s}"!'.format(dvt))
        self.dv_type = dv_type

        # Make sure imageid is always a chunking variable and all chunk vars exist
        if self._fix._imageid not in chunks:
            chunks.append(self._fix._imageid)
        for var in chunks:
            if var not in self._fix.variables:
                raise ValueError('Error: chunking variable "{:s}" does not exist in dataset!'.format(var))

        self.chunks = chunks
        self.progress = progress

        # Add any specified features to model
        if features:
            if type(features) != list and type(features) != tuple:
                features = [features,]  # force list of features
            for f in features:
                if feature_labels is not None and len(feature_labels) > 0:
                    label = feature_labels.pop(0)
                elif f.label is not None: 
                    label = f.label
                else:
                    label = self._feat_label(f)
                self.add_feature(f, label=label)

            self.update(progress=self.progress)


    def _feat_label(self, feature):
        """ Create label for unlabeled Feature object."""
        cls = feature.__class__.__name__
        label = 'f' + cls[0:5]

        f_label = label
        suffix = 1

        while f_label in self.features.keys():
            f_label = label + str(suffix)
            suffix += 1
        return f_label


    @property
    def predictors(self):
        """ Model predictor matrix, updated if necessary """
        if not self._consistent:
            self.update()
        return self._pred


    def __repr__(self):
        """ String representation for print summary. """
        if not self._consistent:
            self.update()

        r = '<gridfix.FixationModel, {:d} samples, DV={:s}, chunked by: {:s}>\n'.format(self.predictors.shape[0], str(self.dv_type), str(self.chunks))
        r += 'Fixations:\n\t{:s}\n'.format(str(self._fix))
        r += 'Regions:\n\t{:s}\n'.format(str(self.regionset))

        if len(self.features) > 0:
            r += '\nFeatures:\n'
            for l,f in self.features.items():
                r += '\t{:s}\t{:s}\n'.format(l, f.__class__.__name__)

        if len(self.comp_features) > 0:
            r += '\nFeature Comparisons:\n'
            for l,f in self.comp_features.items():
                r += '{:s}:\n'.format(l)
                for code, feat in f.items():
                    r += '\t{:s}\t{:s}\n'.format(str(code), feat.__class__.__name__)
        return(r)


    def r_source(self, datafile='gridfix.csv', comments=True, scale=True, center=True, 
                 optimizer=None, fixed=None, random=None, random_slopes=False):
        """ Generate R source code from current feature settings.

        Args:
            datafile (str): predictor matrix file name (for R import via read.table())
            comments (boolean): if True, add explanatory comments and headers to source
            scale (boolean): if True, add code to scale (normalize) feature predictors
            center (boolean): if True, add code to center (demean) feature predictors
            optimizer (str): optional optimizer to pass to R glmer()
            fixed (list): list of column names (strings) to add as fixed factors
            random (list): list of column names (strings) to add as random factors
            random_slopes (boolean): also add random slopes to generated R code

        Returns:
            R source code as string
        """
        r_libs = ['lme4']

        src = ''
        if comments:
            d = time.strftime('%d.%m.%y, %H:%M:%S', time.localtime())
            src =  '# GridFix GLMM R source, generated on {:s}\n# \n'.format(d)
            src += '# Predictor file:\t{:s}\n'.format(datafile)
            src += '# Fixations file:\t{:s}\n'.format(str(self._fix.input_file))
            src += '# RegionSet:\t\t{:s}\n'.format(str(self.regionset))
            src += '# DV type(s):\t\t{:s}\n'.format(str(self.dv_type))
            src += '\n'

        # Libraries
        for lib in r_libs:
            src += 'library({:s})\n'.format(lib)
        src += '\n'

        # Predictor file
        src += 'gridfixdata  <- read.table("{:s}", header=T, sep="\\t", row.names=NULL)\n\n'.format(datafile)

        # Factors
        if comments:
            src += '# Define R factors for all chunking variables and group dummy codes\n'
        for chunk in self.chunks:
            src += 'gridfixdata${:s} <- as.factor(gridfixdata${:s})\n'.format(chunk, chunk)
        if len(self.comp_features) > 0:
            for cf in self.comp_features.keys():
                src += 'gridfixdata${:s} <- as.factor(gridfixdata${:s})\n'.format(cf, cf)
        src += '\n'

        # Center and scale
        if scale or center:
            if len(self.features) > 0:
                r_cent = 'FALSE'
                r_scal = 'FALSE'
                if center:
                    r_cent = 'TRUE'
                if scale:
                    r_scal = 'TRUE'
                if comments:
                    src += '# Center and scale predictors\n'
                for f in self.features.keys():
                    src += 'gridfixdata${:s}_C <- scale(gridfixdata${:s}, center={:s}, scale={:s})\n'.format(f, f, r_cent, r_scal)
                src += '\n'

        # GLMM model formula (DV is set later)
        formula = '{:s} ~ 1'

        if fixed is None:
            # Best guess: all simple features should be fixed factors!
            if len(self.features) > 0:
                fixed = self.features.keys()
            else:
                fixed = []

        fixed_vars = ''
        for f in fixed:
            try:
                # likely a feature object
                fl = f.label
            except AttributeError:
                # text label specified
                fl = f
            if scale or center:
                fixed_vars += ' + {:s}_C '.format(fl)
            else:
                fixed_vars += ' + {:s} '.format(fl)
        formula += fixed_vars

        if random is None:
            # imageid should be a random factor by default
            random = [self._fix._imageid]
        
        if len(fixed) > 0:
            for r in random:
                try:
                # likely a feature object
                    rl = r.label
                except AttributeError:
                    rl = r
                if random_slopes:
                    formula += ' + (1{:s} | {:s})'.format(fixed_vars, rl)
                else:
                    formula += ' + (1 | {:s})'.format(rl)

        # Optimizer parameter
        opt_call = ''
        if optimizer is not None:
            opt_call = ', control=glmerControl(optimizer="{:s}")'.format(optimizer)

        if comments:
            src += '# NOTE: this source code can only serve as a scaffolding for your own analysis!\n'
            src += '# You MUST adapt the GLMM model formula below to your model, then uncomment the corresponding line!\n'

        # GLMM model call(s) - one per requested DV
        models = []
        for current_dv in self.dv_type:
            if self._dvs[current_dv]['fun'] is None:
                # Don't generate R code for a DV with no model function specified
                continue

            models.append('model.{:s}'.format(current_dv))

            if comments:
                src += '# DV: {:s}\n#'.format(current_dv)

            src += 'model.{:s} <- {:s}({:s}, data=gridfixdata{:s}'.format(current_dv, self._dvs[current_dv]['fun'], formula.format(self._dvs[current_dv]['rvar']), opt_call)
            if self._dvs[current_dv]['family'] is not None:
                src += ', family={:s}'.format(self._dvs[current_dv]['family'])
            src += ')\n\n'

        out_f, ext = os.path.splitext(datafile)
        r_objlist = ','.join(['"{:s}"'.format(a) for a in models])
        src += 'save(file="{}_GLMM.Rdata", list = c({:s}))\n\n'.format(out_f, r_objlist)
        src += 'print(summary(model))\n'
        return src


    def _process_chunk(self, chunk_vals, data, pred_columns, group_levels):
        """ Process a single data chunk. """
        if data is not None:
            sel_labels = dict(zip(self.chunks, chunk_vals))
            imageid = str(sel_labels[self._fix._imageid])
            tmpdf = DataFrame(columns=pred_columns, index=range(len(self.regionset[imageid])))

            # groupby returns a single string if only one chunk columns is selected
            if type(chunk_vals) != tuple:
                chunk_vals = (chunk_vals,)

            # Fixations
            subset = self._fix.select_fix(sel_labels)

            # Chunk and region values
            for col in self.chunks:
                tmpdf[col] = data[col].iloc[0]

            # Region ID and numbering
            if self.regionset.is_global:
                tmpdf.regionid = np.array(self.regionset.info[self.regionset.info.imageid == '*'].regionid, dtype=str)
                tmpdf.regionno = np.array(self.regionset.info[self.regionset.info.imageid == '*'].regionno, dtype=int)
            else:
                tmpdf.regionid = np.array(self.regionset.info[self.regionset.info.imageid == imageid].regionid, dtype=str)
                tmpdf.regionno = np.array(self.regionset.info[self.regionset.info.imageid == imageid].regionno, dtype=int)

            # Dependent Variables (DVs)
            for dv in self._dvs.keys():
                if dv in self.dv_type:
                    if self._dvs[dv]['f'] == 'c':
                        # Fixation count based measures
                        tmpdf[self._dvs[dv]['rvar']] = self.regionset.fixated(subset, var=dv, imageid=imageid, exclude_first=self.exclude_first_fix,
                                                                                  exclude_last=self.exclude_last_fix)

                    if self._dvs[dv]['f'] == 't' and subset.has_times:
                        # Fixation time based measures
                        tmpdf[self._dvs[dv]['rvar']] = self.regionset.fixtimes(subset, var=dv, imageid=imageid, exclude_first=self.exclude_first_fix,
                                                                                  exclude_last=self.exclude_last_fix)

            # Simple per-image features
            for feat_col, feat in self.features.items():
                tmpdf[feat_col] = feat.apply(imageid, normalize=self.normalize_features)

            # Feature group comparisons
            if len(self.comp_features) > 0:
                for levels in group_levels:
                    for idx, gc in enumerate(self.comp_features.keys()):
                        tmpdf[gc] = levels[idx]
                        tmpdf['{:s}_val'.format(gc)] = self.comp_features[gc][levels[idx]].apply(imageid, normalize=self.normalize_features)

            return tmpdf


    def update(self, progress=False):
        """ Update predictor matrix from features (this may take a while).

        Args:
            progress (boolean): if True, print model creation progress
        """
        ts = time.time()

        # Output DF columns
        pred_columns = self.chunks + ['regionid', 'regionno', 'dvFix']
        pred_columns += list(self.features.keys())
        for cf in self.comp_features.keys():
            pred_columns += [cf, '{:s}_val'.format(cf)]

        pred_new = DataFrame(columns=pred_columns)

        splitdata = self._fix.data.groupby(self.chunks)

        group_levels = []
        if len(self.comp_features) > 0:
            groups = [list(f.keys()) for i,f in self.comp_features.items()]
            group_levels = list(itertools.product(*groups))

        # Process individual chunks
        for chunk_vals, data in splitdata:
            if progress:
                print(chunk_vals)
            results = self._process_chunk(chunk_vals, data, pred_columns, group_levels)
            pred_new = pred_new.append(results)

        self._pred = pred_new
        self._consistent = True
        self.runtime = time.time() - ts     # rebuild duration
        

    def add_feature(self, feature, label=None):
        """ Add a feature to the model.

        Args:
            feature (Feature): Feature object to add
            label (str): label and output column name for this feature
        """
        # Generate unique feature label
        if label is None:
            label = self._feat_label(feature)

        # Check feature length
        if len(feature) != len(self.regionset):
            w = 'Could not add feature "{:s}": invalid length ({:d} instead of {:d})!'
            raise ValueError(label, len(feature), len(self.regionset))

        self.features[label] = feature
        self._consistent = False


    def add_comparison(self, features, codes=None, label=None):
        """ Add a feature comparison group to the model.

        This generates a long-style predictor matrix for the specified features,
        needed to compare e.g. different saliency maps in their relative effects.

        Args:
            features (list): list of Feature objects to combine into a group
            codes (list): numeric codes to use in "dummy coding", e.g. [0, 1, 2]
            label (str): label and output column name for this feature group
        """
        # Generate unique group label
        if label is None:
            suffix = 1
            g_label = 'fC' + str(suffix)

            while g_label in self.comp_features.keys():
                suffix += 1
                g_label = 'fC' + str(suffix)

        # Generate dummy codes if necessary
        if codes is None:
            codes = range(0, len(features))
        else:
            codes = [int(c) for c in codes]

        comp = {codes[c]: f for c,f in enumerate(features)}
        self.comp_features[g_label] = comp
        self._consistent = False


    def save(self, basename, sep='\t', pred=True, pred_pickle=False, src=True, src_comments=True, precision=10,
             optimizer=None, fixed=None, random=None, random_slopes=False):
        """ Saves the predictor matrix to a CSV text file.

        Args:
            basename (str): base filename to save to, without extension
            sep (str): item separator, default TAB
            pred (boolean): if True, output predictor matrix as CSV
            pred_pickle (boolean): if True, also save predictors to Pickle object
            src (boolean): if True, output r source code file for lme4
            src_comments (boolean): if True, add comments to source code
            precision (int): number of decimal places for CSV (default: 10)
            optimizer (str): optional optimizer to pass to R glmer()
            fixed (list): list of column names (strings) to add as fixed factors
            random (list): list of column names (strings) to add as random factors
            random_slopes (boolean): also add random slopes to generated R code

        """
        if not self._consistent:
            self.update()

        if pred:
            if LooseVersion(pandas_version) >= LooseVersion('0.17.1'):
                # compression supported from 0.17.1
                f_pred = '{:s}.csv.gz'.format(basename)
                self.predictors.to_csv(f_pred, sep, index=False, float_format='%.{:d}f'.format(precision), compression='gzip')
            else:
                f_pred = '{:s}.csv'.format(basename)
                self.predictors.to_csv(f_pred, sep, index=False, float_format='%.{:d}f'.format(precision))

        if pred_pickle:
            f_pred = '{:s}.pkl'.format(basename)
            self.predictors.to_pickle(f_pred)

        if src:
            f_src = '{:s}.R'.format(basename)
            src = self.r_source(comments=src_comments, datafile=f_pred, optimizer=optimizer,
                                fixed=fixed, random=random, random_slopes=random_slopes)
            with open(f_src, 'w') as sf:
                sf.write(src)


if __name__ == '__main__':
    print('The gridfix modules cannot be called directly. Please use one of the included tools, e.g. gridmap.')



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os

import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

from PIL import Image

from fractions import Fraction
from pandas import DataFrame, read_table

from .model import ImageSet


class RegionSet(object):
    """ Base class for sets of image regions of interest.

    RegionSets can be used to spatially group fixations, create Feature objects
    for a FixationModel and split an image into parts. Classes inheriting from 
    RegionSet may specify functions to create regions.

    Attributes:
        info (DataFrame): table of region metadata (labels, bboxes, number of pixels...)
        imageids (list): list of all imageids associated with this RegionSet
        is_global (bool): True if regions are global (non-image-specific)
        label (str): optional label to distinguish between RegionSets.
        memory_usage (float): memory usage of all binary masks (kiB)
        size (tuple): image dimensions, specified as (width, height).
    """

    def __init__(self, size, regions, region_labels=None, label=None, add_background=False):
        """ Create a new RegionSet from existing region masks.

        Args:
            size (tuple): image dimensions, specified as (width, height)
            regions: 3d ndarray (bool) with global set of masks 
                OR dict of multiple such ndarrays, with imageids as keys
            region_labels: list of region labels IF _regions_ is a single array
                OR dict of such lists, with imageids as keys
            label (str): optional descriptive label for this RegionSet
            add_background (bool): if True, this creates a special region to capture all
                fixations that don't fall on an explicit region ("background" fixations)

        Raises:
            ValueError if incorrectly formatted regions/region_labels provided
        """
        self._regions = {'*': np.ndarray((0,0,0))}
        self._labels  = {'*': []}

        self.size = size
        self.label = label
        self._msize = (size[1], size[0])   # matrix convention
        self.has_background = False

        if isinstance(regions, dict):
            # Dict with image-specific region ndarrays
            self._regions = regions
            if region_labels is not None and isinstance(region_labels, dict) and len(regions) == len(region_labels):
                # Check imageids for consistency
                for r in regions.keys():
                    if r not in region_labels.keys():
                        raise ValueError('Labels not consistent: {:s} not in region_labels'.format(r))
                for r in region_labels.keys():
                    if r not in regions.keys():
                        raise ValueError('Labels not consistent: {:s} not in regions'.format(r))
                self._labels = region_labels
            else:
                self._labels = {}
                for imid in regions:
                    self._labels[imid] = [str(x+1) for x in range(len(regions[imid]))]

        elif isinstance(regions, np.ndarray):
            # Single array of regions - assume global region set ('*')
            if regions.shape[1:] == self._msize:
                self._regions['*'] = regions.astype(bool)
            if region_labels is not None and len(region_labels) == regions.shape[0]:
                self._labels['*'] = region_labels
            else:
                self._labels['*'] = [str(x+1) for x in range(regions.shape[0])]

        else:
            raise ValueError('First argument for RegionSet creation must be ndarray ' + 
                             '(global regions) or dict of ndarrays (image-specific regions)!')

        if add_background:
            for iid in self._regions.keys():
                bgmask = ~self.mask(iid).reshape(1, size[1], size[0])
                self._regions[iid] = np.concatenate([self._regions[iid], bgmask], axis=0)
                self._labels[iid].append('__BG__')
            self.has_background = True

        self.info = self._region_metadata()


    def __repr__(self):
        """ String representation """
        r = 'gridfix.RegionSet(label={:s}, size=({:d}, {:d}),\nregions={:s},\nregion_labels={:s})'
        return r.format(str(self.label), self.size[0], self.size[1], str(self._regions), str(self._labels))


    def __str__(self):
        """ Short string representation for printing """
        r = '<{:s}{:s}, size={:s}, {:d} region{:s}{:s}, memory={:.1f} kB>'
        
        myclass = str(self.__class__.__name__)
        if self.label is not None:
            lab = ' ({:s})'.format(self.label)
        else:
            lab = ''

        num_s = ''
        num_r = len(self)
        if num_r > 1:
            num_s = 's'

        imid_s = ''
        if len(self._regions) > 1 and not self.is_global:
            imid_s = ' in {:d} images'.format(len(self._regions))

        return r.format(myclass, lab, str(self.size), num_r, num_s, imid_s, self.memory_usage)


    def __len__(self):
        """ Overload len(RegionSet) to report total number of regions. """
        if self.is_global:                
            return len(self._regions['*'])
        else:
            num_r = 0
            for imid in self._regions:
                num_r += len(self._regions[imid])
            return num_r


    def __getitem__(self, imageid):
        """ Bracket indexing returns all region masks for a specified imageid.
            If global regions are set ('*'), always return global region set. 
        """
        return self._select_region(imageid)


    def _region_metadata(self):
        """ Return DataFrame of region metadata """
        info_cols = ['imageid', 'regionid', 'regionno', 'left', 'top', 'right', 'bottom', 'width', 'height', 'area', 'imgfrac']
        info = []

        if self.is_global:
            imageids = ['*']
        else:
            imageids = self.imageids

        for imid in imageids:
            reg = self._select_region(imid)
            lab = self._select_labels(imid)

            for i,l in enumerate(lab):
                a = np.argwhere(reg[i])
                if a.shape[0] > 0:
                    (top, left) = a.min(0)[0:2]
                    (bottom, right) = a.max(0)[0:2]
                    (width, height) = (right-left+1, bottom-top+1)
                    area = reg[i][reg[i] > 0].sum()
                    imgfrac = round(area / (reg[i].shape[0] * reg[i].shape[1]), 4)
                else:
                    # Region is empty - shouldn't, but can happen with add_background at full coverage
                    (top, left, bottom, right, width, height, area, imgfrac) = (0,) * 8

                rmeta = [imid, l, i+1, left, top, right, bottom, width, height, area, imgfrac]
                info.append(rmeta)

        return DataFrame(info, columns=info_cols)


    def _select_region(self, imageid=None):
        """ Select region by imageid with consistency check """
        if self.is_global:
            return(self._regions['*'])

        if imageid is not None and imageid in self._regions.keys():
            return(self._regions[imageid])
        else:
            raise ValueError('RegionSet contains image-specific regions, but no valid imageid was specified!')


    def _select_labels(self, imageid=None):
        """ Select region labels corresponding to _select_region """
        if self.is_global:
            return(self._labels['*'])

        if imageid is not None and imageid in self._regions.keys():
            return(self._labels[imageid])
        else:
            raise ValueError('RegionSet contains image-specific regions, but no valid imageid was specified!')


    @property
    def is_global(self):
        """ Return True if a global map is defined (key '*') """
        if '*' in self._regions.keys():
            return True
        else:
            return False


    @property
    def imageids(self):
        """ Return list of imageids for which region maps exist """
        if self.is_global:
            return []
        
        imids = []
        for imid in self._regions.keys():
            imids.append(imid)
        return imids


    @property
    def memory_usage(self):
        """ Calculate size in memory of all regions combined """
        msize = 0.0
        for reg in self._regions.keys():
            msize += float(self._regions[reg].nbytes) / 1024.0
        return msize


    def count_map(self, imageid=None, ignore_background=True):
        """ Return the number of regions referencing each pixel.

        Args:
            imageid (str):  if set, return map for specified image only 
            ignore_background (bool): if True, ignore auto-generated background region

        Returns:
            2d ndarray of image size, counting number of regions for each pixel
        """
        
        cm = np.zeros(self._msize, dtype=int)

        if self.is_global:
            for reidx, re in enumerate(self._regions['*'][:, ...]):
                if ignore_background and self._labels['*'][reidx] == '__BG__':
                    continue
                cm += re.astype(int)
            return cm

        elif imageid is None:
            for imid in self._regions:
                if imid == '*':
                    continue
                for reidx, re in enumerate(self._regions[imid][:, ...]):
                    if ignore_background and self._labels[imid][reidx] == '__BG__':
                        continue
                    cm += re.astype(int)

        else:
            r = self._select_region(imageid)
            l = self._select_labels(imageid)
            for reidx, re in enumerate(r[:, ...]):
                if ignore_background and l[reidx] == '__BG__':
                    continue
                cm += re.astype(int)

        return cm

    
    def mask(self, imageid=None, ignore_background=True):
        """ Return union mask of all regions or regions for specified image.

        Args:
            imageid (str):  if set, return mask for specified image only
            ignore_background (bool): if True, ignore auto-generated background region

        Returns:
            2d ndarray of image size (bool), True where at least one region
            references the corresponding pixel.
        """
        return self.count_map(imageid, ignore_background).astype(bool)


    def region_map(self, imageid=None, ignore_background=True):
        """ Return map of region numbers, global or image-specifid.

        Args:
            imageid (str):  if set, return map for specified image only
            ignore_background (bool): if True, ignore auto-generated background region

        Returns:
            2d ndarray (int), containing the number (ascending) of the last
            region referencing the corresponding pixel.
        """
        apply_regions = self._select_region(imageid)
        apply_labels = self._select_labels(imageid)
        tmpmap = np.zeros(self._msize)
        for idx, region in enumerate(apply_regions):
            if ignore_background and apply_labels[idx] == '__BG__':
                continue
            tmpmap[region] = (idx + 1)
        return tmpmap


    def coverage(self, imageid=None, normalize=False, ignore_background=True):
        """ Calculates coverage of the total image size as a scalar.

        Args:
            imageid (str):  if set, return coverage for specified image only 
            normalize (bool): if True, divide global result by number of imageids in set.
            ignore_background (bool): if True, ignore auto-generated background region

        Returns:
            Total coverage as a floating point number.
        """
        if imageid is not None:
            counts = self.count_map(imageid, ignore_background)
            cov = float(counts.sum()) / float(self.size[0] * self.size[1])
            return cov
        else:
            # Global coverage for all imageids
            cm = np.zeros(self._msize, dtype=int)
            for re in self._regions.keys():
                if re == '*':
                    cm += self.count_map('*', ignore_background)
                    break
                cm += self.count_map(re, ignore_background)

            cov = float(cm.sum()) / float(self.size[0] * self.size[1])
            if normalize:
                cov = cov / len(self)
            return cov


    def plot(self, imageid=None, values=None, cmap=None, image_only=False, ax=None, alpha=1.0):
        """ Plot regions as map of shaded areas with/without corresponding feature values

        Args:
            imageid (str): if set, plot regions for specified image
            values (array-like): one feature value per region
            cmap (str): name of matplotlib colormap to use to distinguish regions
            image_only (boolean): if True, return only image content without axes
            ax (Axes): axes object to draw to, to include result in other figure
            alpha (float): opacity of plotted regions (set < 1 to visualize overlap)
        
        Returns:
            matplotlib figure object, or None if passed an axis to draw on
        """
        apply_regions = self._select_region(imageid)
        tmpmap = np.zeros(self._msize)

        if ax is not None:
            ax1 = ax
        else:
            fig = plt.figure()
            ax1 = fig.add_subplot(1,1,1)

        if cmap is None:
            if values is None and 'viridis' in plt.colormaps():
                cmap = 'viridis'
            else:
                cmap = 'gray'

        if type(cmap) == str:
            cmap = plt.get_cmap(cmap)

        if alpha < 1.0:
            # allow stacking by setting masked values transparent
            alpha_cmap = cmap
            alpha_cmap.set_bad(alpha=0)

            ax1.imshow(tmpmap, cmap=plt.get_cmap('gray'), interpolation='none')

            for idx, region in enumerate(apply_regions):
                rmap = np.zeros(self._msize)
                if values is not None and len(values) == apply_regions.shape[0]:
                    rmap[region] = values[idx]
                    ax1.imshow(np.ma.masked_equal(rmap, 0), cmap=alpha_cmap, interpolation='none', alpha=alpha, 
                               vmin=0, vmax=np.nanmax(values))

                else:
                    rmap[region] = idx + 1
                    ax1.imshow(np.ma.masked_equal(rmap, 0), cmap=alpha_cmap, interpolation='none', alpha=alpha, 
                               vmin=0, vmax=apply_regions.shape[0])

        else:
            # If no alpha requested, this is much faster but doesn't show overlap
            ax1.imshow(tmpmap, cmap=plt.get_cmap('gray'), interpolation='none')
            
            if values is not None and len(values) == apply_regions.shape[0]:
                rmap = np.zeros(self._msize)
                for idx, region in enumerate(apply_regions):
                    rmap[region] = values[idx]
                ax1.imshow(np.ma.masked_equal(rmap, 0), cmap=cmap, interpolation='none', vmin=0, vmax=np.nanmax(values))
            else:
                ax1.imshow(np.ma.masked_equal(self.region_map(imageid), 0), cmap=cmap, interpolation='none',
                           vmin=0, vmax=apply_regions.shape[0])

        if image_only:
            ax1.axis('off')
        
        else:
            t = '{:s}'.format(self.__class__.__name__)
            if self.label is not None:
                t += ': {:s}'.format(self.label)
            if imageid is not None:
                t += ' (img: {:s})'.format(imageid)
            ax1.set_title(t)

        if ax is None and not plt.isinteractive():  # see ImageSet.plot()
            return fig


    def plot_regions_on_image(self, imageid=None, imageset=None, image_cmap=None, cmap=None, plotcolor=None, 
                              fill=False, alpha=0.4, labels=False, image_only=False, ax=None):
        """ Plot region bounding boxes on corresponding image

        Args:
            imageid (str): if set, plot regions for specified image
            imageset (ImageSet): ImageSet object containing background image/map
            image_cmap (str): name of matplotlib colormap to use for image
            cmap (str): name of matplotlib colormap to use for bounding boxes
            plotcolor (color): matplotlib color for bboxes (overrides colormap)
            fill (boolean): draw shaded filled rectangles instead of boxes
            alpha (float): rectangle opacity (only when fill=True)
            labels (boolean): if True, draw text labels next to regions
            image_only (boolean): if True, return only image content without axes
            ax (Axes): axes object to draw to, to include result in other figure

        Returns:
            matplotlib figure object, or None if passed an axis to draw on
        """
        if imageset is None or imageid not in imageset.imageids:
            raise ValueError('To plot regions on top of image, specify ImageSet containing corresponding background image!')

        if ax is not None:
            ax1 = ax
        else:
            fig = plt.figure()
            ax1 = fig.add_subplot(1,1,1)

        if image_cmap is not None:
            if type(image_cmap) == str:
                image_cmap = plt.get_cmap(image_cmap)
            ax1.imshow(imageset[imageid], cmap=image_cmap, interpolation='none')
        else:
            ax1.imshow(imageset[imageid], interpolation='none')

        if cmap is None:
            if 'viridis' in plt.colormaps():
                cmap = 'viridis'
            else:
                cmap = 'hsv'

        if type(cmap) == str:
            boxcolors = plt.get_cmap(cmap)
        else:
            boxcolors = cmap
        cstep = 0

        if self.is_global:
            rmeta = self.info[self.info.imageid == '*']
        else:
            rmeta = self.info[self.info.imageid == imageid]

        for idx, region in rmeta.iterrows():
            if self.has_background and region.regionid == '__BG__':
                # Always skip background region when drawing bboxes
                continue
            if plotcolor is None:
                c = boxcolors(cstep/len(rmeta))
            else:
                c = plotcolor
            cstep += 1
            if not fill:
                ax1.add_patch(Rectangle((region.left, region.top), region.width, region.height, color=c, fill=False, linewidth=2))
            else:
                ax1.add_patch(Rectangle((region.left, region.top), region.width, region.height, color=c, linewidth=0, alpha=0.7))
            if labels:
                # Draw text labels with sensible default positions
                if region.right > (self.size[0] * .95):
                    tx = region.right
                    ha = 'right'
                else:
                    tx = region.left
                    ha = 'left'
                if region.bottom > (self.size[1] * .95):
                    ty = region.top - 5
                else:
                    ty = region.bottom + 20
                ax1.text(tx, ty, region.regionid, horizontalalignment=ha)

        if image_only:
            ax1.axis('off')

        else:
            t = '{:s}'.format(self.__class__.__name__)
            if self.label is not None:
                t += ': {:s}'.format(self.label)
            t += ' (img: {:s})'.format(imageid)
            ax1.set_title(t)

        if ax is None and not plt.isinteractive():  # see ImageSet.plot()
            return fig


    def apply(self, image, imageid=None, crop=False, ignore_background=True):
        """ Apply this RegionSet to a specified image.

        Returns a list of the image arrays "cut out" by each region mask, with
        non-selected image areas in black. If regionset is not global, _imageid_ needs
        to be specified!

        Args:
            image (ndarray): image array to be segmented.
            imageid (str): valid imageid (to select image-specific regions if not a global regionset)
            crop (bool): if True, return image cropped to bounding box of selected area
            ignore_background (bool): if True, ignore auto-generated background region

        Returns:
            If crop=False, a list of ndarrays of same size as image, with non-selected areas
            zeroed. Else a list of image patch arrays cropped to bounding box size.
        """
        slices = []
        apply_regions = self._select_region(imageid)
        apply_labels = self._select_labels(imageid)

        for idx, region in enumerate(apply_regions):
            if ignore_background and apply_labels[idx] == '__BG__':
                continue
            mask = (region == True)
            out = np.zeros(image.shape)
            out[mask] = image[mask]

            if crop:
                a = np.argwhere(out)
                if a.shape[0] > 0:
                    (ul_x, ul_y) = a.min(0)[0:2]
                    (br_x, br_y) = a.max(0)[0:2]
                    out = out[ul_x:br_x+1, ul_y:br_y+1]
            slices.append(out)

        return slices


    def export_patches(self, image, imageid=None, crop=True, image_format='png',
                       rescale=False, ignore_background=True):
        """ Apply this RegionSet to an image array and save the resulting image patches as files.

        Saves an image of each image part "cut out" by each region mask, cropped by default.
        If the RegionSet is not global, imageid needs to be specified!

        Args:
            image (ndarray): image array to be segmented.
            imageid (str): imageid (to select image-specific regions if not a global regionset)
            crop (bool): if True, return image cropped to bounding box of selected area
            image_format (str): image format that PIL understands (will also be used for extension)
            rescale (bool): if True, scale pixel values to full 0..255 range
                before saving (e.g., for saliency maps)
            ignore_background (bool): if True, ignore auto-generated background region

        """
        apply_regions = self._select_region(imageid)
        apply_labels = self._select_labels(imageid)
        imstr = '{:s}_{:s}.{:s}'

        for idx, region in enumerate(apply_regions):
            if ignore_background and apply_labels[idx] == '__BG__':
                continue

            mask = (region == True)
            out = np.zeros(image.shape)
            out[mask] = image[mask]

            if crop:
                a = np.argwhere(out)
                if a.shape[0] > 0:
                    (ul_x, ul_y) = a.min(0)[0:2]
                    (br_x, br_y) = a.max(0)[0:2]
                    out = out[ul_x:br_x+1, ul_y:br_y+1]

            if imageid is None or imageid == '*':
                imageid = 'image'

            if rescale:
                out = (out - out.min()) / out.max() * 255.0
            else:
                out *= 255.0

            rimg = Image.fromarray(np.array(out, np.uint8))
            rimg.save(imstr.format(str(imageid), str(apply_labels[idx]), image_format), image_format)


    def export_patches_from_set(self, imageset, crop=True, image_format='png', rescale=False, ignore_background=True):
        """ Save all sliced image patches from an ImageSet as image files. 

        Saves an image of each image part "cut out" by each region mask, cropped by default.
        If the RegionSet is not global, only images with valid region masks will be processed.

        Args:
            imageset (ImageSet): a valid ImageSet containing images to slice
            imageid (str): imageid (to select image-specific regions if not a global regionset)
            crop (bool): if True, return image cropped to bounding box of selected area
            image_format (str): image format that PIL understands (will also be used for extension)
            rescale (bool): if True, scale pixel values to full 0..255 range
                before saving (e.g., for saliency maps)
            ignore_background (bool): if True, ignore auto-generated background region

        """
        if not isinstance(imageset, ImageSet):
            raise TypeError('First argument must be an ImageSet! To slice a single image, use export_patches().')

        for cimg in imageset.imageids:
            if not self.is_global and cimg not in self.imageids:
                print('Warning: RegionSet contains image-specific regions, but no regions available for {:s}. Skipped.'.format(cimg))
            else:
                self.export_patches(imageset[cimg], imageid=cimg, crop=crop, image_format=image_format,
                                    rescale=rescale, ignore_background=ignore_background)
            

    def fixated(self, fixations, var='fixated', imageid=None, exclude_first=False, exclude_last=False):
        """ Returns visited / fixated regions using data from a Fixations object.

        Args:
            fixations (Fixations/DataFrame): fixation data to test against regions
            var (str): type of fixation mapping variable to calculate (default: 'fixated'):
                'fixated': fixation status: 0 - region was not fixated, 1 - fixated (default)
                'count': total number of fixations on each region
                'fixid': fixation ID (from input dataset) for first fixation in each region
            imageid (str): imageid (to select image-specific regions if not a global regionset)
            exclude_first (bool): if True, first fixated region will always be returned as NaN
            exclude_last (str): controls how to deal with regions receiving the last image fixation:
                'never' or False: do not handle the last fixation specially
                'always' or True: drop the entire region if it received the last fixation at any time
                'pass': exclude viewing pass (one or multiple fixations) that received the last fixation

        Returns:
            1D ndarray (float) containing number of fixations per region (if count=True) 
            or the values 0.0 (region was not fixated) or 1.0 (region was fixated)
        """
        if type(exclude_last) == bool:
            if exclude_last:
                exclude_last = 'always'
            elif not exclude_last:
                exclude_last = 'never'

        apply_regions = self._select_region(imageid)
        vis = np.zeros(apply_regions.shape[0], dtype=float)

        # Drop out-of-bounds fixations
        fix = fixations.data[(fixations.data[fixations._xpx] >= 0) & 
                             (fixations.data[fixations._xpx] < self.size[0]) & 
                             (fixations.data[fixations._ypx] >= 0) & 
                             (fixations.data[fixations._ypx] < self.size[1])]
    
        if len(fix) > 0:
            if exclude_first:
                first_fix = fixations.data[fixations.data[fixations._fixid] == min(fixations.data[fixations._fixid])]
                if len(first_fix) > 1:
                    print('Warning: you have requested to drop the first fixated region, but more than one ' +
                          'location ({:d}) matches the lowest fixation ID! Either your fixation ' .format(len(first_fix)) +
                          'IDs are not unique or the passed dataset contains data from multiple images or conditions.')
            if exclude_last != 'never':
                last_fix = fixations.data[fixations.data[fixations._fixid] == max(fixations.data[fixations._fixid])]
                if len(last_fix) > 1:
                    print('Warning: you have requested to drop the last fixated region, but more than one ' +
                          'location ({:d}) matches the highest fixation ID! Either your fixation ' .format(len(last_fix)) +
                          'IDs are not unique or the passed dataset contains data from multiple images or conditions.')

            for (idx, roi) in enumerate(apply_regions):
                if exclude_first:
                    try:
                        is_first = roi[first_fix[fixations._ypx], first_fix[fixations._xpx]]
                        if isinstance(is_first, np.ndarray) and np.any(is_first):
                            vis[idx] = np.nan
                            continue
                        elif is_first:
                            vis[idx] = np.nan
                            continue
                    except IndexError:
                        pass # last fixation is out of bounds for image!

                if exclude_last == 'always':
                    try:
                        is_last = roi[last_fix[fixations._ypx], last_fix[fixations._xpx]]
                        if isinstance(is_last, np.ndarray) and np.any(is_last):
                            vis[idx] = np.nan
                            continue
                        elif is_last:
                            vis[idx] = np.nan
                            continue
                    except IndexError:
                        pass # last fixation is out of bounds for image!

                fv = roi[fix[fixations._ypx], fix[fixations._xpx]]
                if np.any(fv):
                    rfix = fix[fv] # All fixations on region
                    bystart = rfix[rfix[fixations._fixstart] >= 0].sort_values(fixations._fixid)

                    if len(bystart) > 0:
                        # Find viewing passes (sets of in-region fixations without leaving region)
                        idxvalid = np.ones(bystart.shape[0], dtype=np.bool) # fix indices to keep
                        idxdiff = bystart[fixations._fixid].diff().reset_index(drop=True)
                        pass_onsets = idxdiff.index.values[(idxdiff > 1)].tolist()
                        num_refix = len(pass_onsets)
                        num_passes = num_refix + 1
                        if len(pass_onsets) >= 1:
                            end_first_pass = pass_onsets[0]
                        else:
                            end_first_pass = bystart.shape[0]

                        # If requested, remove pass containing the last fixation
                        if exclude_last == 'pass':
                            passes = [0,] + pass_onsets + [len(bystart)+1,]
                            for pidx in range(0, len(passes)-1):
                                passfix = bystart.iloc[passes[pidx]:passes[pidx+1], :]
                                if last_fix.index.values[0] in passfix.index:
                                    # Exclude this and all following passes. Note that no later passes
                                    # should exist unless there is an index error in the fixation data!
                                    idxvalid[passes[pidx]:] = False
                                    break

                            if np.all(idxvalid == False):
                                # If no valid fixations remain, drop the whole region (NA)
                                vis[idx] = np.nan
                                continue
                            else:
                                # Keep only valid fixations for fixation count measures
                                bystart = bystart.loc[idxvalid, :]
                                num_refix = np.sum(idxvalid[pass_onsets] == True)
                                num_passes = num_refix + 1

                        # Calculate fixation status measures
                        if var == 'count':
                            # Number of fixations in region
                            vis[idx] = bystart.shape[0]

                        elif var == 'fixated':
                            # Binary coding of fixation status
                            vis[idx] = (bystart.shape[0] >= 1.0)

                        elif var == 'fixid':
                            # Return first valid fixation ID in region
                            if bystart.shape[0] >= 1.0:
                                vis[idx] = bystart.loc[bystart.index[0], fixations._fixid]
                            else:
                                vis[idx] = np.nan

                        elif var == 'passes':
                            # Total number of fixation passes
                            vis[idx] = num_passes

                        elif var == 'refix':
                            # Total number of fixation passes
                            vis[idx] = num_refix

                else:
                    # No fixations in region -> fixID should be NA
                    if var == 'fixid':
                        vis[idx] = np.nan

        return vis


    def fixtimes(self, fixations, var='total', imageid=None, exclude_first=False, exclude_last=False):
        """ Returns fixation-timing based variable for each region. Default is total viewing time.

        Args:
            fixations (Fixations/DataFrame): fixation data to test against regions
            var (str): type of fixation time variable to calculate (default: 'total'):
                'total': total fixation time for each region
                'gaze': gaze duration, i.e. total fixation time in first pass
                'first': first fixation duration per region
                'single': fixation duration if region was fixated exactly once
                'tofirst': start time of the first fixation on each region
            imageid (str): imageid (to select image-specific regions if not a global regionset)
            exclude_first (bool): if True, first fixated region will always be returned as NaN
            exclude_last (str): controls how to deal with regions receiving the last image fixation:
                'never' or False: do not handle the last fixation specially
                'always' or True: drop the entire region if it received the last fixation at any time
                'pass': exclude viewing pass (one or multiple fixations) that received the last fixation

        Returns:
            1D ndarray (float) containing fixation time based dependent variable for each region.
            Regions that were never fixated according to criteria will be returned as NaN.
        """
        if var not in ['total', 'gaze', 'first', 'single', 'tofirst']:
            raise ValueError('Unknown fixation time variable specified: {:s}'.format(var))

        if not fixations.has_times:
            raise AttributeError('Trying to extract a time-based DV from a dataset without fixation timing information! Specify fixstart=/fixend= when loading fixation data!')

        if type(exclude_last) == bool:
            if exclude_last:
                exclude_last = 'always'
            elif not exclude_last:
                exclude_last = 'never'

        apply_regions = self._select_region(imageid)
        ft = np.ones(apply_regions.shape[0], dtype=float) * np.nan

        # Drop out-of-bounds fixations
        fix = fixations.data[(fixations.data[fixations._xpx] >= 0) &
                             (fixations.data[fixations._xpx] < self.size[0]) &
                             (fixations.data[fixations._ypx] >= 0) &
                             (fixations.data[fixations._ypx] < self.size[1])]

        if len(fix) > 0:
            if exclude_first:
                first_fix = fixations.data[fixations.data[fixations._fixid] == min(fixations.data[fixations._fixid])]
                if len(first_fix) > 1:
                    print('Warning: you have requested to drop the first fixated region, but more than one ' +
                          'location ({:d}) matches the lowest fixation ID! Either your fixation ' .format(len(first_fix)) +
                          'IDs are not unique or the passed dataset contains data from multiple images or conditions.')

            if exclude_last != 'never':
                last_fix = fixations.data[fixations.data[fixations._fixid] == max(fixations.data[fixations._fixid])]
                if len(last_fix) > 1:
                    print('Warning: you have requested to drop the last fixated region, but more than one ' +
                          'location ({:d}) matches the highest fixation ID! Either your fixation ' .format(len(last_fix)) +
                          'IDs are not unique or the passed dataset contains data from multiple images or conditions.')

            for (idx, roi) in enumerate(apply_regions):
                if exclude_first:
                    try:
                        is_first = roi[first_fix[fixations._ypx], first_fix[fixations._xpx]]
                        if isinstance(is_first, np.ndarray) and np.any(is_first):
                            ft[idx] = np.nan
                            continue
                        elif is_first:
                            ft[idx] = np.nan
                            continue
                    except IndexError:
                        pass # first fixation is out of bounds for image!

                if exclude_last == 'always':
                    # If this region has the last fixation, drop it here (NaN) and move on
                    try:
                        is_last = roi[last_fix[fixations._ypx], last_fix[fixations._xpx]]
                        if isinstance(is_last, np.ndarray) and np.any(is_last):
                            ft[idx] = np.nan
                            continue
                        elif is_last:
                            ft[idx] = np.nan
                            continue
                    except IndexError:
                        pass # last fixation is out of bounds for image!

                fidx = roi[fix[fixations._ypx], fix[fixations._xpx]]
                if np.any(fidx):
                    rfix = fix[fidx]    # all fixations in this region
                    bystart = rfix[rfix[fixations._fixstart] >= 0].sort_values(fixations._fixid)

                    if len(bystart) > 0:
                        # Find viewing passes (sets of in-region fixations without leaving region)
                        idxvalid = np.ones(bystart.shape[0], dtype=np.bool) # fix indices to keep
                        idxdiff = bystart[fixations._fixid].diff().reset_index(drop=True)
                        pass_onsets = idxdiff.index.values[(idxdiff > 1)].tolist()
                        num_refix = len(pass_onsets)
                        num_passes = num_refix + 1
                        if len(pass_onsets) >= 1:
                            end_first_pass = pass_onsets[0]
                        else:
                            end_first_pass = bystart.shape[0]

                        # If requested, remove pass containing the last fixation
                        if exclude_last == 'pass':
                            passes = [0,] + pass_onsets + [len(bystart)+1,]
                            for pidx in range(0, len(passes)-1):
                                passfix = bystart.iloc[passes[pidx]:passes[pidx+1], :]
                                if last_fix.index.values[0] in passfix.index:
                                    # Exclude this and all following passes. Note that no later passes
                                    # should exist unless there is an index error in the fixation data!
                                    idxvalid[passes[pidx]:] = False
                                    break

                            if np.all(idxvalid == False):
                                # If no valid fixations remain, drop the whole region (NA)
                                ft[idx] = np.nan
                                continue
                            else:
                                # Keep only valid fixations for fixation count measures
                                bystart = bystart.loc[idxvalid, :]
                                num_refix = np.sum(idxvalid[pass_onsets] == True)
                                num_passes = num_refix + 1

                        # Calculate fixation timing measures
                        if var == 'gaze':
                            # Gaze duration: total viewing time of first pass only
                            ft[idx] = sum(bystart.loc[bystart.index[0:end_first_pass], fixations._fixdur])

                        elif var == 'first':
                            # First fixation duration
                            ft[idx] = bystart.loc[bystart.index[0], fixations._fixdur]

                        elif var == 'single':
                            # Single fixation duration (=first fixation duration if not refixated in first pass)
                            ft[idx] = bystart.loc[bystart.index[0], fixations._fixdur]

                            # If refixated on first pass, set to NaN instead
                            if end_first_pass > 1:
                                ft[idx] = np.nan

                        elif var == 'tofirst':
                            # Time until first fixation / first fixation onset
                            ft[idx] = bystart.loc[bystart.index[0], fixations._fixstart]

                        elif var == 'total':
                            # Total viewing time of valid fixations
                            ft[idx] = sum(bystart.loc[:, fixations._fixdur])

        return ft



class GridRegionSet(RegionSet):
    """ RegionSet defining an n-by-m regular grid covering the full image size.

    Attributes:
        cells (list): list of bounding box tuples for each cell,
            each formatted as (left, top, right, bottom)
        gridsize (tuple): grid dimensions as (width, height). If unspecified,
            gridfix will try to choose a sensible default.
        label (string): optional label to distinguish between RegionSets
    """

    def __init__(self, size, gridsize=None, label=None, region_labels=None):
        """ Create a new grid RegionSet

        Args:
            size (tuple): image dimensions, specified as (width, height).
            gridsize(tuple): grid dimensions, specified as (width, height).
            region_labels (string): list of optional region labels (default: cell#)
        """

        if gridsize is None:
            gridsize = self._suggest_grid(size)
            print('Note: no grid size was specified. Using {:d}x{:d} based on image size.'.format(gridsize[0], gridsize[1]))

        (regions, cells) = self._grid(size, gridsize)
        RegionSet.__init__(self, size=size, regions=regions, label=label, region_labels=region_labels,
                           add_background=False) # GridRegionSets are exhaustive, so the 'background' is empty.

        self.gridsize = gridsize

        # List of region bboxes
        self.cells = cells


    def __str__(self):
        """ Short string representation for printing """
        r = '<gridfix.GridRegionSet{:s}, size={:s}, {:d}x{:d} grid, {:d} cell{:s}, memory={:.1f} kB>'
        
        if self.label is not None:
            lab = ' ({:s})'.format(self.label)
        else:
            lab = ''

        num_s = ''
        num_r = len(self)
        if num_r > 1:
            num_s = 's'
        return r.format(lab, str(self.size), self.gridsize[0], self.gridsize[1], num_r, 
                        num_s, self.memory_usage)


    def _suggest_grid(self, size):
        """ Suggest grid dimensions based on image size.

        Args:
            size (tuple): image dimensions, specified as (width, height).

        Returns:
            Suggested grid size tuple as (width, height).
        """
        aspect = Fraction(size[0], size[1])
        s_width = aspect.numerator
        s_height = aspect.denominator
        if s_width < 6:
            s_width *= 2
            s_height *= 2
        return (s_width, s_height)


    def _grid(self, size, gridsize):
        """ Build m-by-n (width,height) grid as 3D nparray.

        Args:
            size (tuple): image dimensions, specified as (width, height).
            gridsize(tuple): grid dimensions, specified as (width, height).

        Returns:
            tuple containing the grid regions and their bounding box coordinates
            as (grid, cells):

                grid (numpy.ndarray): regions for RegionSet creation
                cells (list): list of bounding box tuples for each cell, 
                    each formatted as (left, top, right, bottom)

        """
        (width, height) = size
        _msize = (size[1], size[0])
        cell_x = int(width / gridsize[0])
        cell_y = int(height / gridsize[1])
        n_cells = int(gridsize[0] * gridsize[1])

        grid = np.zeros((n_cells,) + _msize, dtype=bool)
        cells = []

        # Sanity check: do nothing if image dimensions not cleanly divisible by grid
        if width % gridsize[0] > 0 or height % gridsize[1] > 0:
            e = 'Error: image dimensions not cleanly divisible by grid! image=({:d}x{:d}), grid=({:d}x{:d})'
            raise ValueError(e.format(width, height, gridsize[0], gridsize[1]))

        # Create a mask of 1s/True for each cell
        cellno = 0
        for y_es in range(0, height, cell_y):
            for x_es in range(0, width, cell_x):
                mask = np.zeros(_msize, dtype=bool)
                mask[y_es:y_es + cell_y, x_es:x_es + cell_x] = True
                grid[cellno,...] = mask
                cells.append((x_es, y_es, x_es + cell_x, y_es + cell_y))
                cellno += 1

        return (grid, cells)



class BBoxRegionSet(RegionSet):
    """ RegionSet based on rectangular bounding boxes.

    Attributes:
        cells (list): list of bounding box tuples for each cell,
            each formatted as (left, top, right, bottom)
        label (string): optional label to distinguish between RegionSets
        from_file (string): filename in case regions were loaded from file
        padding (tuple): padding in pixels as ('left', 'top', 'right', 'bottom')
    """

    def __init__(self, size, bounding_boxes, label=None, region_labels=None, sep='\t',
                 imageid='imageid', regionid='regionid', bbox_cols=('x1', 'y1', 'x2', 'y2'),
                 padding=0, add_background=False, coord_format=None):
        """ Create new BBoxRegionSet

        Args:
            size (tuple):   image dimensions, specified as (width, height).
            bounding_boxes: one of the following:
                name of a text/CSV file with columns ([imageid], [regionid], x1, y1, x2, y2)
                list of 4-tuples OR 2D ndarray with columns (x1, y1, x2, y2) for global bboxes
            region_labels (str): list of optional region labels if bounding_boxes is a global array/list
            imageid (str): name of imageid column in input file (if not present, bboxes will be treated as global)
            regionid (str): name of regionid column in input file
            sep (str): separator to use when reading files
            bbox_cols: tuple of column names for ('left', 'top', 'right', 'bottom')
            padding (int): optional bbox padding in pixels as ('left', 'top', 'right', 'bottom'),
                or a single integer to specify equal padding on all sides
            add_background (bool): if True, this creates a special region to capture all
                fixations that don't fall on an explicit region ("background" fixations)
            coord_format (str): Defines how input x and y coordinates are interpreted:
                'oneindexed': coordinates start at 1, e.g. 1..100 for a 100px box
                'zeroindexed': coordinates start at 0, e.g. 0..99 for a 100px box
                'apple': coordinates start at 0, but end at <size>, e.g. 0..100 for a 100px box,
                         in this convention, the pixels sit "between" coordinate values
        """
        self.input_file = None
        self.input_df = None

        self._imageid = imageid
        self._regionid = regionid
        self._cols = bbox_cols

        if coord_format is None:
            err = 'No coordinate format specified! Please provide coord_format argument:\n'
            err += '"oneindexed": coordinates start at 1, e.g. 1..100 for a 100px box\n'
            err += '"zeroindexed": coordinates start at 0, e.g. 0..99 for a 100px box\n'
            err += '"apple": coordinates start at 0, but end at <size>, e.g. 0..100 for a 100px box.'
            raise ValueError(err)

        if type(padding) == int:
            self.padding = (padding,) * 4
        else:
            self.padding = padding

        if isinstance(bounding_boxes, DataFrame):
            # Passed a DataFrame
            bbox = bounding_boxes

        elif type(bounding_boxes) == str:
            # Passed a file name
            try:
                bbox = read_table(bounding_boxes, sep=sep)
                self.input_file = bounding_boxes

            except:
                raise ValueError('String argument supplied to BBoxRegionSet, but not a valid CSV file!')

        else:
            # Try array type
            try:
                bbox = DataFrame(bounding_boxes, columns=['x1', 'y1', 'x2', 'y2'])
            except:
                raise ValueError('Supplied argument to BBoxRegionSet not in the form (x1, y1, x2, y2)')

        (regions, labels) = self._parse_bbox_df(bbox, size, padding, coord_format=coord_format)
        if region_labels is not None:
            labels = region_labels

        RegionSet.__init__(self, size=size, regions=regions, label=label, region_labels=labels, add_background=add_background)

        self.input_df = bbox


    def _parse_bbox_df(self, df, size, padding, coord_format='oneindexed'):
        """ Parse a DataFrame of bounding boxes into a region dict.

        Args:
            df (DataFrame): DataFrame of bounding box coordinates
            size (tuple): image size as (width, height), to check for out-of-bounds coordinates
            padding (int): padding in pixels as ('left', 'top', 'right', 'bottom'), default (0,0,0,0)
            coord_format (str): Defines how input x and y coordinates are interpreted (see __init__)

        Returns:
            tuple of dicts as (regions, labels), using imageids as keys. Resulting dicts can be
            passed to RegionSet.__init__ directly.
        """
        regions = {}
        labels = {}
        _msize = (size[1], size[0])

        if self._imageid in df.columns:
            # Image-specific bounding boxes

            # Force imageid to strings
            df[self._imageid] = df[self._imageid].astype(str)

            for imid, block in df.groupby(self._imageid):
                N = block.shape[0]
                reg = np.zeros((N,) + _msize, dtype=bool)
                lab = []
                bidx = 0

                for idx,row in block.iterrows():
                    # left, top, right, bottom
                    c = [row[self._cols[0]], row[self._cols[1]], row[self._cols[2]], row[self._cols[3]]]

                    if self._regionid in df.columns:
                        l = row[self._regionid]
                    else:
                        l = str(bidx + 1)

                    # Convert coordinates to Python indices
                    if coord_format.lower() == 'oneindexed':
                        c = [round(c[0]) - 1,   # Correct lower bounds
                             round(c[1]) - 1,
                             round(c[2]),
                             round(c[3])]
                    elif coord_format.lower() == 'zeroindexed':
                        c = [round(c[0]),
                             round(c[1]),
                             round(c[2]) + 1,   # Python slices are end-excluding,
                             round(c[3]) + 1]   # i.e. [a, b] does not include b
                    elif coord_format.lower() == 'apple':
                        c = [round(c[0]),       # Just round, format is already fine
                             round(c[1]),
                             round(c[2]),
                             round(c[3])]

                    if c[0] > size[0] or c[2] > size[0] or c[1] > size[1] or c[3] > size[1]:
                        err = 'At least one coordinate of region {:s}/{:s} exceeds the specified image size!'
                        raise ValueError(err.format(imid, l))

                    if c[2]-c[0] < 1 or c[3]-c[1] < 1:
                        err = 'At least one dimension of {:s}/{:s} has a negative length! Columns specified in the wrong order?'
                        raise ValueError(err.format(imid, l))

                    # Add padding, ensuring padded bboxes are cropped to image size
                    c = [c[0] - self.padding[0], c[1] - self.padding[1], c[2] + self.padding[2], c[3] + self.padding[3]]
                    if c[0] < 0:
                        c[0] = 0
                    if c[1] < 0:
                        c[1] = 0
                    if c[2] > size[0]:
                        c[2] = size[0]
                    if c[3] > size[1]:
                        c[3] = size[1]

                    mask = np.zeros(_msize, dtype=bool)
                    mask[c[1]:c[3], c[0]:c[2]] = True
                    reg[bidx,...] = mask
                    lab.append(l)
                    bidx += 1

                regions[imid] = reg
                labels[imid] = lab

        else:
            # Global bounding boxes
            reg = reg = np.zeros((df.shape[0], ) + _msize, dtype=bool)
            lab = []

            for idx,row in df.iterrows():
                c = (row[self._cols[0]], row[self._cols[1]], row[self._cols[2]], row[self._cols[3]])

                if self._regionid in df.columns:
                    l = row[self._regionid]
                else:
                    l = str(idx + 1)

                # Convert coordinates to Python indices
                if coord_format.lower() == 'oneindexed':
                    c = [round(c[0]) - 1,   # Correct lower bounds
                         round(c[1]) - 1,
                         round(c[2]),
                         round(c[3])]
                elif coord_format.lower() == 'zeroindexed':
                    c = [round(c[0]),
                         round(c[1]),
                         round(c[2]) + 1,   # Python slices are end-excluding,
                         round(c[3]) + 1]   # i.e. [a, b] does not include b
                elif coord_format.lower() == 'apple':
                    c = [round(c[0]),       # Just round, format is already fine
                         round(c[1]),
                         round(c[2]),
                         round(c[3])]

                if c[0] > size[0] or c[2] > size[0] or c[1] > size[1] or c[3] > size[1]:
                    err = 'At least one coordinate of region {:s} exceeds the specified image size!'
                    raise ValueError(err.format(l))

                mask = np.zeros(_msize, dtype=bool)
                mask[c[1]:c[3], c[0]:c[2]] = True
                reg[idx,...] = mask
                lab.append(l)

            regions = {'*': reg}
            labels = {'*': lab}

        return (regions, labels)


#!/usr/bin/python3
# -*- coding: utf-8 -*-

import types

import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt

from pandas import DataFrame, read_table
from scipy.ndimage import center_of_mass    # for CentralBiasFeature
from scipy.ndimage.filters import sobel     # for SobelEdgeFeature

from .regionset import RegionSet
from .model import ImageSet


class Feature(object):
    """ Base class for image features defined for each region in a set

    A Feature can be described as a two-step "recipe" on a given ImageSet and 
    RegionSet: First, a transform() is executed upon each image array, e.g. a
    filter operation. Second, a combine() operation yields a single value per region.
    The result is combined into a feature vector of len(regionset).

    Attributes:
        regionset (RegionSet): the associated regionset object
        imageset (ImageSet): a set of images or feature maps to process
        length (int): length of feature vector, i.e. number of regions
    """

    def __init__(self, regionset, imageset, trans_fun=None, comb_fun=None, label=None, normalize_output=False):
        """ Create a new basic Feature object.

        Args:
            regionset (RegionSet): a RegionSet object
            trans_fun (function): Function to use for the transformation step instead
                of default. Must accept PIL.Image and return Image or np.ndarray of same size.
            comb_fun (function): Function to use for the reduction step instead of default.
                Must accept a np.ndarray and return a scalar value.
            label (string): optional label to distinguish between Features of the same type
            normalize_output (bool): if True, always normalize output values of this feature to 0..1
        """
        self.regionset = regionset
        self.imageset = imageset
        self.label = label
        self.normalize_output = normalize_output

        self._fvalues = {}

        # Replace default functions only if the specified ones make sense
        try:
            tmp = trans_fun(self, np.array([[0,1],[0,0]]))
            self._transform = self.transform   # keep default around
            self.transform = types.MethodType(trans_fun, self)
        except TypeError:
            if trans_fun is not None:
                print('Warning: trans_fun seems to return invalid values, using default!')

        try:
            tmp = comb_fun(self, np.array([[0,1],[0,0]]), np.array([[0,1],[0,0]]))
            self._combine = self.combine
            self.combine = types.MethodType(comb_fun, self)
        except TypeError:
            if comb_fun is not None:
                print('Warning: comb_fun seems to return invalid values, using default!')


    def __repr__(self):
        """ String representation for print summary. """
        desc = ''
        if self.label is not None:
            desc = ' "{:s}"'.format(str(self.label))

        norm = ''
        if self.normalize_output:
            norm = ', normalized'
        r = '<gridfix.{:s}{:s}, length={:d}{:s}>'.format(self.__class__.__name__, desc, len(self.regionset), norm)
        r += '\nRegions:\n\t{:s}'.format(str(self.regionset))
        r += '\nImages:\n\t{:s}'.format(str(self.imageset))

        return(r)


    def __len__(self):
        """ Overload len() to report length of feature vector. """
        return len(self.regionset)


    def __getitem__(self, imageid):
        """ Bracket indexing using an imageid returns the feature vector. """
        return self.apply(imageid, normalize=False)


    def transform(self, image):
        """ Apply image transform to specified image. 

        In the base Feature class, this simply returns the input image as-is.
        This function may be overloaded in subclasses or replaced by the 
        trans_fun= argument on construction time. 

        Args:
            image (ndarray): image / feature map array to transform

        Returns: 
            np.ndarray of image data
        """
        return np.asarray(image, dtype=float)


    def combine(self, image, region, fun=np.mean):
        """ Combine all selected pixel values in selection using specified function.

        In the base Feature class, this simply returns the mean of all pixels.
        This function may be overloaded in subclasses or replaced by the comb_fun= argument.

        Args:
            image (np.ndarray): 2D feature image
            region (np.ndarray): binary mask array defining a region
            fun (function): function to apply to selection. Must return a scalar.

        Returns:
            Scalar value depending on the specified function.
        """
        return fun(np.asarray(image[region], dtype=float))


    def apply(self, imageid, normalize=None):
        """ Apply feature to a single image from associated ImageSet.

        Args:
            imageid (str): valid ID from associated ImageSet
            normalize (bool): if True, scale output to range 0...1 (default: False)
        Returns:
            1D numpy.ndarray of feature values, same length as regionset
        """
        if imageid not in self._fvalues.keys():
            fv = []
            img_in = self.imageset[imageid]

            # Transformation step
            img_trans = self.transform(img_in)

            # Combination step
            for region in self.regionset[imageid]:
                f = self.combine(np.asarray(img_trans),  # make sure combine gets arrays
                                 np.asarray(region, dtype=bool))
                fv.append(f)

            # cache for later use
            self._fvalues[imageid] = np.array(fv)


        f = self._fvalues[imageid]
        if self.normalize_output:
            # Feature set to always normalize
            return((f - f.min()) / (f.max() - f.min()))
        elif normalize is not None:
            if normalize:
                # Upstream Model requested normalized values
                return((f - f.min()) / (f.max() - f.min()))
        else:
            # Default is to return unmodified values
            return f


    def apply_all(self, normalize=None):
        """ Apply feature to every image in the ImageSet and return a DataFrame.

        Args:
            normalize (bool): if True, scale output to range 0...1 (default: False)

        Returns: DataFrame similar to RegionSet.info with region feature values
        """
        rmeta = self.regionset.info.copy()
        if self.regionset.is_global:
            # If global RegionSet, copy region data once per image
            df = DataFrame(columns=rmeta.columns)
            for img in self.imageset.imageids:
                tmp = rmeta.copy()
                tmp['imageid'] = img
                df = df.append(tmp)
        else:
            df = rmeta
        df['value'] = np.nan

        for img in self.imageset.imageids:
            if img not in self.regionset.imageids and not self.regionset.is_global:
                continue

            img_trans = self.transform(self.imageset[img])
            l = self.regionset._select_labels(img)

            for idx,region in enumerate(self.regionset[img]):
                f = self.combine(np.asarray(img_trans), np.asarray(region, dtype=bool))
                df.loc[(df['imageid'] == img) & (df['region'] == str(l[idx])), 'value'] = f

        # Normalize value column if requested
        if self.normalize_output:
            val = df.loc[:, 'value']
            df.loc[:, 'value'] = (val - val.min()) / (val.max() - val.min())

        elif normalize is not None:
            if normalize:
                val = df.loc[:, 'value']
                df.loc[:, 'value'] = (val - val.min()) / (val.max() - val.min())

        return df


    def plot(self, imageid, what='both', cmap='gray', image_only=False, ax=None):
        """ Display feature map and/or feature values.

        Args:
            imageid (str): valid ID from associated ImageSet
            what (str): show only feature 'values', 'map' or 'both'
            cmap (str): name of matplotlib colormap to use
            image_only (boolean): if True, return only image content without labels
            ax (Axes): axes object to draw to (only if what!='both')

        Returns:
            matplotlib figure object, or None if passed an axis to draw on
        """
        lbl_text = ''
        if self.label is not None:
            lbl_text = ': {:s}'.format(str(self.label))

        if what == 'map':
            if ax is not None:
                ax1 = ax
            else:
                fig = plt.figure()
                ax1 = fig.add_subplot(1,1,1)

            imap = self.transform(self.imageset[imageid])
            ax1.imshow(imap, cmap=plt.get_cmap(cmap), interpolation='none')
            if image_only:
                ax1.axis('off')
            else:
                ax1.set_title('{:s}{:s} (img: {:s})'.format(self.__class__.__name__, lbl_text, imageid))

        elif what == 'values':
            if ax is not None:
                ax1 = ax
            else:
                fig = plt.figure()
                ax1 = fig.add_subplot(1,1,1)

            self.regionset.plot(values=self.apply(imageid), imageid=imageid, cmap=cmap, ax=ax1, image_only=image_only)
            if image_only:
                ax1.axis('off')
            else:
                ax1.set_title('{:s}{:s} (img: {:s})'.format(self.__class__.__name__, lbl_text, imageid))

        else:
            if ax is not None:
                print('Warning: ax argument is only valid if a single plot type is specified! Returning full figure instead.')

            fig = plt.figure()
            ax1 = fig.add_subplot(1,2,1)
            imap = self.transform(self.imageset[imageid])
            ax1.imshow(imap, cmap=plt.get_cmap(cmap), interpolation='none')
            ax2 = fig.add_subplot(1,2,2)
            self.regionset.plot(values=self.apply(imageid), imageid=imageid, cmap=cmap, ax=ax2, image_only=image_only)
            if image_only:
                ax1.axis('off')
                ax2.axis('off')
            else:
                ax1.set_title('map')
                ax2.set_title('values')
                fig.suptitle('{:s}{:s} (img: {:s})'.format(self.__class__.__name__, lbl_text, imageid))

        if ax is None and not plt.isinteractive():  # see ImageSet.plot()
            return fig



class CentralBiasFeature(Feature):
    """ Models central viewer bias as the distance to image center for each region. 

    The exact model of distance depends on the "measure" argument: if set to 'distance'
    (the default), the image transformation step does nothing and euclidean distance is 
    returned. If set to 'gaussian', anisotropic Gaussian distance based on Clarke & Tatler, 
    2014, Vis Res is used and transform() returns the corresponding Gaussian map. 
    """

    def __init__(self, regionset, imageset, measure='gaussian', sig2=0.23, nu=None, label=None, normalize_output=False):
        """ Create a new CentralBiasFeature object.

        Args:
            regionset (RegionSet): region set to apply feature to
            imageset (ImageSet): a set of images or feature maps (optional for this Feature)
            measure (string): distance measure to use ('euclidean', 'gaussian', 'taxicab')
            sig2 (float): variance value for type='gaussian'
            nu (float): anisotropy value for type='gaussian'
            label (string): optional label to distinguish between Features
            normalize_output (bool): if True, always normalize output values of this feature to 0..1
        """
        if measure not in ['gaussian', 'euclidean', 'taxicab']:
            print('Warning: unknown central bias measure "{:s}" specified, falling back to euclidean distance!'.format(measure))
            measure = 'euclidean'

        self.measure = measure
        self._map = None
        self._values = {}

        self.sig2 = sig2
        self.nu = nu

        if nu is None:
            if measure == 'gaussian':
                self.nu = 0.45
            else:
                self.nu = 1
            
        
        def _transform(self, image=None):
            """ For 'gaussian': Gauss distance map, else return empty array. """

            if self._map is None:
                # compute feature map on first call

                mapsize = self.imageset.size[1], self.imageset.size[0]

                if measure == 'gaussian':
                    self._map = self._aniso_gauss(mapsize, self.sig2, self.nu)
                else:
                    self._map = np.zeros(mapsize, dtype=float)

            return self._map
            

        def _combine(self, image, region):
            """ Calculate distance from region center of mass to image center. """
            
            mapsize = self.imageset.size

            com = np.round(center_of_mass(region))
            img_center = (int(round(mapsize[1] / 2)), int(round(mapsize[0] / 2)))

            if self.measure == 'gaussian':
                return image[int(com[0]), int(com[1])]
        
            elif self.measure == 'euclidean':
                dist = np.hypot((com[0] - img_center[0]) / self.nu, com[1] - img_center[1])
                return np.round(dist, 0)

            elif measure == 'taxicab':
                dist = abs(com[0] - img_center[0]) + abs(com[1] - img_center[1])
                return np.round(dist, 0)

        self._trans_fun = _transform
        self._comb_fun = _combine

        Feature.__init__(self, regionset, imageset, trans_fun=_transform, comb_fun=_combine, label=label,
                         normalize_output=normalize_output)


    def apply(self, imageid=None, normalize=None):
        """ Apply central bias to image, returning region distance values.

        Args:
            imageid (str): for consistency, ignored for central bias (same for all images)
            normalize (boolean): if True, scale output to range 0...1 (default: False)

        Returns:
            1D numpy.ndarray of feature values, same length as regionset
        """
        if imageid not in self._values.keys():
            fv = []
            
            img_trans = self.transform()

            for region in self.regionset[imageid]:
                f = self.combine(img_trans, np.asarray(region, dtype=bool))
                fv.append(f)

            self._values[imageid] = np.array(fv)

        f = self._values[imageid]
        if self.normalize_output:
            # Feature set to always normalize
            return((f - f.min()) / (f.max() - f.min()))
        elif normalize is not None:
            if normalize:
                # Upstream Model requested normalized values
                return((f - f.min()) / (f.max() - f.min()))
        else:
            # Default is to return unmodified values
            return f


    def __repr__(self):
        """ String representation for printing """
        desc = ''
        if self.label is not None:
            desc = ' "{:s}"'.format(str(self.label))

        r = '<gridfix.CentralBiasFeature{:s}, length={:d}, measure "{:s}"'.format(desc, len(self.regionset), self.measure)

        if self.measure == 'gaussian':
            r += ', sig2={:.2f}, nu={:.2f}'.format(self.sig2, self.nu)

        r += '>'
        r += '\nRegions:\n\t{:s}'.format(str(self.regionset))
        r += '\nImages:\n\t{:s}'.format(str(self.imageset))
        return(r)


    def _aniso_gauss(self, shape, sig2=0.23, nu=0.45):
        """ Calculate anisotropic, aspect-corrected Gaussian central bias map.

        This function yields an image-sized Gaussian map of the distance
        from image center, including anisotropy as described in Clarke &
        Tatler, 2014, Vis Res.

        Args:
            shape (tuple): image ndarray shape (height, width)
            sig2 (float): variance of Gaussian
            nu (float): anisotropy coefficient

        Returns:
            2D ndarray containing Gaussian central distance map
            (values increase with larger distance from center)
        """
        height, width = shape[0], shape[1]
        ar = width / height
        x = np.linspace(-1, 1, width)
        y = np.linspace(-1 / ar , 1 / ar, height)
        (xx, yy) = np.meshgrid(x, y)

        G = np.exp(-((xx ** 2) / sig2 + (yy ** 2) / sig2 / nu) / 2)
        return 1 - G



class LuminanceFeature(Feature):
    """ Models mean image luminance in each region """

    def __init__(self, regionset, imageset, label=None, normalize_output=False):
        """ Create a new LuminanceFeature
        
        Args:
            regionset: a RegionSet to be evaluated
            imageset: ImageSet containing images or feature maps to process
            label (str): optional label to distinguish between Features
            normalize_output (bool): if True, always normalize output values of this feature to 0..1

        """

        def _transform(self, image):
            """ Convert 3D-RGB image to 2D-intensity (like rgb2gray.m). """
            if image.ndim > 2 and image.shape[2] == 3:
                R, G, B = image[:, :, 0], image[:, :, 1], image[:, :, 2]
                lum = 0.2989 * R + 0.5870 * G + 0.1140 * B
                return lum
            else:
                return image

        def _combine(self, image, region):
            """ Return mean region luminance. """
            return(image[region].mean())

        self.trans_fun = _transform
        self.comb_fun = _combine

        Feature.__init__(self, regionset, imageset, trans_fun=_transform, comb_fun=_combine, label=label,
                         normalize_output=normalize_output)



class LumContrastFeature(Feature):
    """ Feature based on local luminance contrast in each region """

    def __init__(self, regionset, imageset, label=None, normalize_output=False):
        """ Create a new LuminanceContrastFeature

        Args:
            regionset: a RegionSet to be evaluated
            imageset: ImageSet containing images or feature maps to process
            label (str): optional label to distinguish between Features
            normalize_output (bool): if True, always normalize output values of this feature to 0..1
        """

        def _transform(self, image):
            """ Convert 3D-RGB image to 2D-intensity (like rgb2gray.m). """
            if image.ndim > 2 and image.shape[2] == 3:
                R, G, B = image[:, :, 0], image[:, :, 1], image[:, :, 2]
                lum = 0.2989 * R + 0.5870 * G + 0.1140 * B
                return lum
            else:
                return image

        def _combine(self, image, region):
            """ Return local contrast of luminance image. """
            return(image[region].std() / image.mean())

        self.trans_fun = _transform
        self.comb_fun = _combine

        Feature.__init__(self, regionset, imageset, trans_fun=_transform, comb_fun=_combine, label=label,
                         normalize_output=normalize_output)



class SobelEdgeFeature(Feature):
    """ Feature based on relative prevalence of edges within each region """

    def __init__(self, regionset, imageset, label=None, normalize_output=False):
        """ Create a new SobelEdgeFeature

        Args:
            regionset: a RegionSet to be evaluated
            imageset: ImageSet containing images or feature maps to process
            label (str): optional label to distinguish between Features
            normalize_output (bool): if True, always normalize output values of this feature to 0..1
        """

        def _transform(self, image):
            """ Run sobel filter on luminance image, then binarize. """
            if image.ndim > 2 and image.shape[2] == 3:
                R, G, B = image[:, :, 0], image[:, :, 1], image[:, :, 2]
                lum = 0.2989 * R + 0.5870 * G + 0.1140 * B
            else:
                lum = image

            sx = sobel(lum, 0)
            sy = sobel(lum, 1)
            si = np.hypot(sx, sy)

            # Determine threshold for edge detection (adapted from MATLAB edge.m)
            scale = 4.0
            cutoff = scale * si.mean()
            thresh = np.sqrt(cutoff)

            return np.asarray(si >= thresh, dtype=np.uint8)


        def _combine(self, image, region):
            """ Mean of binary image yields fraction of edges. """
            return(image[region].mean())

        self.trans_fun = _transform
        self.comb_fun = _combine

        Feature.__init__(self, regionset, imageset, trans_fun=_transform, comb_fun=_combine, label=label,
                         normalize_output=normalize_output)



class MapFeature(Feature):
    """ Feature to apply a statistical function to each region in feature maps

    Attributes:
        stat (function): the statistics function to apply to each region
    """

    def __init__(self, regionset, imageset, stat=np.mean, label=None, normalize_output=False):
        """ Create a new MapFeature

        Args:
            regionset: a RegionSet to be evaluated
            imageset: ImageSet containing images or feature maps to process
            stat (function): the statistics function to apply to each region
            label (str): optional label to distinguish between Features
            normalize_output (bool): if True, always normalize output values of this feature to 0..1
        """
        self.stat = stat

        def _transform(self, image):
            """ Does nothing, expects predefined input map! """
            return image

        def _combine(self, image, region):
            """ Applies stat function to region. """
            return(self.stat(image[region]))

        self.trans_fun = _transform
        self.comb_fun = _combine

        Feature.__init__(self, regionset, imageset, trans_fun=_transform, comb_fun=_combine, label=label,
                         normalize_output=normalize_output)



class RegionFeature(Feature):
    """ Feature based on properties of the regions in a RegionSet. Allows to use
        e.g. region area or fraction of image pixels as a model predictor.

    Attributes:
        region_property (str): column from DataFrame to return as region feature
    """

    def __init__(self, regionset, imageset, region_property='area', label=None, normalize_output=False):
        """ Create a new RegionFeature

        Args:
            regionset: a RegionSet to be evaluated
            imageset: ImageSet containing images or feature maps to process
            region_property (str): column from DataFrame to return as region feature
            label (str): optional label to distinguish between Features
            normalize_output (bool): if True, always normalize output values of this feature to 0..1
        """
        if region_property not in regionset.info.columns:
            raise ValueError('Specified region property is not a column in RegionSet.info DataFrame! Example: "area"')
        self.region_property = region_property

        def _transform(self, image):
            """ Does nothing """
            return image

        def _combine(self, image, region):
            """ Does nothing """
            return None

        self.trans_fun = _transform
        self.comb_fun = _combine

        Feature.__init__(self, regionset, imageset, trans_fun=_transform, comb_fun=_combine, label=label,
                         normalize_output=normalize_output)


    def apply(self, imageid, normalize=None):
        """ Return selected region property of each region for specified imageid.

        Args:
            imageid (str): valid ID from associated ImageSet
            normalize (bool): if True, scale output to range 0...1 (default: False)

        Returns:
            1D numpy.ndarray of feature values, same length as regionset
        """
        if self.regionset.is_global:
            imageid = '*'
        if imageid not in self.regionset._regions.keys():
            raise ValueError('The imageid specified for RegionFeature was not found in the associated Imageset!')

        sel_df = self.regionset.info[self.regionset.info.imageid == imageid]

        f = np.array(sel_df[self.region_property])
        if self.normalize_output:
            # Feature set to always normalize
            return((f - f.min()) / (f.max() - f.min()))
        elif normalize is not None:
            if normalize:
                # Upstream Model requested normalized values
                return((f - f.min()) / (f.max() - f.min()))
        else:
            # Default is to return unmodified values
            return f






import setuptools

with open('README.md', 'r') as rf:
    readme = rf.read()

setuptools.setup(
    name='gridfix',
    version='0.3.1',
    description='Parcellation-based gaze fixation preprocessing for GLMM analysis',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/ischtz/gridfix',
    author='Immo Schuetz',
    author_email='schuetz.immo@gmail.com',
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'matplotlib', 'scipy', 'pandas'],
    zip_safe=True,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.6')

